"""

"""
import subprocess
import time
import sys
import bisect
import inspect
import os
from dataclasses import dataclass
from functools import wraps, lru_cache
from typing import Any, Dict, Union, Optional, Tuple, Sequence, Mapping
import random

import numpy as np
import torch
from torch import distributed as dist
from torch import nn
from torch.optim import Optimizer
from torch.utils.data import DataLoader
from functools import lru_cache
from .environ import globs
from .experiment import TrainerExperiment
from .logger import Logger
from .saver import Saver
from .meter import Meter, AvgMeter
from .rnd import RndManager
from .datamodule import DataModule
from .mixin import ModelMix, CallbackMix, DataModuleMix

from .params import ParamsType
from lumo.base_classes import attr, TrainerStage
from lumo.base_classes.metaclasses import Merge
from ..proc.const import TRAINER
from lumo.utils import dist
from lumo.utils.device import construct_device_args_kwargs, to_device_enumrate
from lumo.contrib.accelerate import Accelerator


@dataclass()
class initial_tuple():
    models: bool = False
    callbacks: bool = False
    optims: bool = False
    train_dataloader: bool = False
    val_dataloader: bool = False
    test_dataloader: bool = False


class TrainerResult(attr):
    MSG_OK = 'success'
    MSG_OK_MID = 'success stop in midway'
    MSG_NO_DATALOADER = 'dataloader not loaded'

    def __init__(self, state: TrainerStage, meter: AvgMeter, msg=''):
        super().__init__()
        self.state = state
        self.meter = meter
        self.msg = msg


def mp_agent(rank, trainer, op, dataloader):
    import torch.distributed as dist
    trainer.params.local_rank = rank
    dist.init_process_group(backend='nccl', init_method=trainer.params.init_method,
                            rank=rank,
                            world_size=trainer.params.world_size)

    trainer.params.device = 'cuda:{}'.format(rank)
    trainer.regist_device(torch.device(trainer.params.device))
    torch.cuda.set_device(trainer.params.local_rank)
    print('in rank {}'.format(rank))
    op(dataloader)


class _BaseTrainer(ModelMix, CallbackMix, metaclass=Merge):
    """
    1. 组装所有的插件
    2. 提供完整的训练流程 api
    3. 提供对需要 api 函数的 callback


    Trainer helps you to control all stage happened in deeplearning, including train/evaluation/test, save/load checkpoints
    , log/meter/tensorboard variable, etc. It's also very convenient to custom your own logits.

    Thanks for git, trainer will commit all change on `experiments` branch before you start your training, and
    this makes you easily



    When create a Trainer instance, a Params instance should be passed
    """

    __exp_name__ = None
    _call_backs = {
        "save_keypoint", "save_checkpoint", "save_model", "load_state_dict",
        'ioptims', 'icallbacks', 'imodels',
    }

    def _exit_hook(self, exc_type, exc, tb, *args):
        import traceback
        res = traceback.format_exception(exc_type, exc, tb)
        res = [i for i in res if 'in _newfunc' not in i]
        print(''.join(res), file=sys.stderr)

    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)
        if cls.__exp_name__ is None:
            cls.__exp_name__ = cls.__name__.lower().replace("trainer", "Exp")

        from lumo.utils.exithook import replace

        replace(self._exit_hook)

        def wrapper(func, _call_set: list):
            """
            对每个 Trainer 的 _call_backs 类变量中定义的函数尝试绑定回调
            Args:
                func:
                _call_set:

            Returns:

            """

            @wraps(func)
            def _newfunc(*aargs, **kkwargs):
                """执行前回调 on_begin() 、执行后回调 on_end()、执行异常则回调 on_exception() """
                for callback in _call_set:
                    callback.on_begin(self, func, self.params, *aargs, **kkwargs)
                try:
                    _meter = func(*aargs, **kkwargs)
                except BaseException as e:
                    _handles = [callback.on_exception(self, func, self.params, e, *aargs, **kkwargs)
                                for callback in _call_set]

                    if any(_handles):
                        return None
                    else:
                        raise e

                for callback in _call_set:
                    callback.on_end(self, func, self.params, _meter, *aargs, **kkwargs)
                return _meter

            return _newfunc

        self._callback_set = []
        self._callback_name_set = set()

        vars = dir(self)
        for name in vars:
            if name not in self._call_backs:
                continue
            if name.startswith("_"):
                continue
            value = getattr(self, name, None)
            if value is None:
                continue
            if callable(value):
                setattr(self, name, wrapper(value, self._callback_set))
        return self

    @classmethod
    def _gene_class_exp_name(trainer_instance) -> str:
        try:
            file = inspect.getfile(trainer_instance)
            pre = os.path.splitext(os.path.basename(file))[0]
        except:
            pre = 'builtin'

        return "{}.{}".format(pre.lower(), trainer_instance.__exp_name__.lower())

    def __init__(self, params: ParamsType):
        self._state_dicts = attr({
            "models": {},
            "optims": {},
            "buffer": {
                'np': {},
                'th': {}
            },
            "others": {},
            'device': torch.device('cpu'),
            'params': params,
            'exp': TrainerExperiment(self._gene_class_exp_name()),
        })
        self._initialize_globs()
        self._initial_exp()

        self._logger = None
        self._rnd = None
        self._accelerate = None
        self._saver = None
        self._datamodule = None

        self.initial = initial_tuple()
        self.train_epoch_toggle = False
        self.train_toggle = False

        if 'device' in params:
            _device = torch.device(params.device)
            self.regist_device(_device)

        self._check_cb_init()
        self.exp.start()

    def __setattr__(self, name: str, value: Any) -> None:
        super().__setattr__(name, value)
        from torch.optim.optimizer import Optimizer

        if name.startswith('_') or name.endswith('_'):
            # when callback is trainer itself, _hooked will be passed, and caused recursion
            # if some pretrained models need to be ignored in save/load stage, it can be named like 'some_' or '_some'
            return

        if isinstance(value, torch.device):
            self._state_dicts[TRAINER.DKEY.devices][name] = value
        elif isinstance(value, torch.nn.Module):
            self._state_dicts[TRAINER.DKEY.models][name] = value
        elif isinstance(value, Optimizer):
            self._state_dicts[TRAINER.DKEY.optims][name] = value
        elif isinstance(value, (torch.Tensor)):
            self._state_dicts[TRAINER.DKEY.tensor]['th'][name] = value
        elif isinstance(value, (np.ndarray)):
            self._state_dicts[TRAINER.DKEY.tensor]['np'][name] = value
        elif callable(getattr(value, "state_dict", None)) and callable(getattr(value, "load_state_dict", None)):
            self._state_dicts[TRAINER.DKEY.others][name] = value

    def __setitem__(self, key: str, value: Any):
        self.__setattr__(key, value)

    def __setstate__(self, state):
        self._state_dicts = attr.from_dict(state)

    def __getstate__(self):
        res = self._state_dicts.pickify()
        return res

    def _initialize_globs(self):
        globs['rank'] = dist.local_rank()
        globs['world_size'] = dist.world_size()
        globs['LOCAL_RANK'] = globs['rank']
        globs['WORLD_SIZE'] = dist.world_size()

        for k, v in self.params.items():  # type:str,Any
            if k.isupper():
                globs[k] = v
                if isinstance(v, str):
                    os.environ[k] = v
        for k in {'nocommit'}:
            if k in self.params:
                globs[k] = self.params[k]

    def _initial_exp(self):
        self.params.to_json(self.exp.params_fn)

    @staticmethod
    def _iter_dataloader(dataloader, device):
        pass

    def _check_optim_init(self):
        if not self.initial.optims:
            self.ioptims(self.params)
            self.initial.optims = True

    def _check_models_init(self):
        if not self.initial.models:
            self.imodels(self.params)
            self.initial.models = True

    def _check_cb_init(self):
        if not self.initial.callbacks:
            self.icallbacks(self.params)
            self.initial.callbacks = True

    def get_state(self, key, default=None):
        return self._state_dicts.get(key, default)

    def regist_state(self, key, value=None, picklable=True):
        if picklable:
            self._state_dicts[key] = value

    @property
    def global_step(self):
        return self.params.global_step

    @property
    def eidx(self):
        return self.params.eidx

    @property
    def is_dist(self):
        return dist.is_dist()

    @property
    def local_rank(self):
        return dist.local_rank()

    @property
    def world_size(self):
        return dist.world_size()

    @property
    def exp(self) -> TrainerExperiment:
        return self._state_dicts[TRAINER.DKEY.exp]

    @property
    def logger(self):
        if self._logger is None:
            from lumo.kit.logger import set_global_logger
            self._logger = Logger()
            set_global_logger(self._logger)
            if self.params.get('debug', False):
                self._logger.set_verbose(Logger.V_DEBUG)
                self._logger.dddebug('Enable debug log.')
            fn = self._logger.add_log_dir(self.exp.log_dir)
            self.exp.writeline('logger', fn)
        return self._logger

    @property
    def params(self) -> ParamsType:
        return self._state_dicts[TRAINER.DKEY.params]

    @property
    def accelerator(self) -> Accelerator:
        if self._accelerate is None:
            acce = {'device_placement': True,
                    'split_batches': False,
                    'fp16': None,
                    'cpu': False,
                    'rng_types': None,
                    'kwargs_handlers': None,
                    'device': self.params.get('device', None)}
            for k in acce:
                if k in self.params:
                    acce[k] = self.params[k]

            self._accelerate = Accelerator(**acce)
        return self._accelerate

    @property
    def rnd(self) -> RndManager:
        if self._rnd is None:
            self._rnd = RndManager(self.exp.rnd_dir)
        return self._rnd

    @property
    def saver(self) -> Saver:
        if self._saver is None:
            self._saver = Saver(self.exp.saver_dir)
        return self._saver

    @property
    def device(self) -> torch.device:
        return self._state_dicts[TRAINER.DKEY.device]

    @property
    def device_arg_kwargs(self) -> Tuple[Sequence, dict]:
        return construct_device_args_kwargs(self.device)

    @property
    def _models(self):
        return self._state_dicts[TRAINER.DKEY.models]

    @property
    def _optims(self):
        return self._state_dicts[TRAINER.DKEY.optims]

    @property
    def model_dict(self) -> Dict[str, nn.Module]:
        self._check_models_init()
        return self._models

    @property
    def optim_dict(self) -> Dict[str, Optimizer]:
        self._check_optim_init()
        return self._state_dicts[TRAINER.DKEY.optims]

    @property
    def buffers(self) -> Dict[str, Dict[str, Union[np.ndarray, torch.Tensor]]]:
        return self._state_dicts[TRAINER.DKEY.tensor]

    @property
    def nparr(self) -> Dict[str, np.ndarray]:
        return self._state_dicts[TRAINER.DKEY.tensor]['np']

    @property
    def tharr(self) -> Dict[str, torch.Tensor]:
        return self._state_dicts[TRAINER.DKEY.tensor]['th']

    @property
    def others(self) -> Dict[str, Any]:
        return self._state_dicts[TRAINER.DKEY.others]

    @property
    def safe_writer(self):
        """see trainer.writer"""
        import tensorflow as tf
        import tensorboard as tb
        tf.io.gfile = tb.compat.tensorflow_stub.io.gfile
        return self.writer

    @property
    @lru_cache()
    def writer(self):
        """
        Notes:
        ------
        When using add_embedding, there may raise some exceptions cased by version conflict, here is some solutions:

        1. tensorflow_core._api.v1.io.gfile' or'tensorflow_core._api.v2.io.gfile' has no attribute 'get_filesystem'
        first, try upgrade tensorboard and tensorflow as followed version:
            tensorboard==2.0.2
            tensorflow==2.0.0

        if you still have the same problem, use this code as a temporary solution:

            import tensorflow as tf
            import tensorboard as tb
            tf.io.gfile = tb.compat.tensorflow_stub.io.gfile

        use `trainer.safe_writer` to get a writter with these code added inside thexp.

        solution is referred by https://github.com/pytorch/pytorch/issues/30966


        2. You may cause PermissionError like: [Errno 13] Permission denied: '/tmp/.tensorboard-info/pid-20281.info'
        the solution is to set environment variable TMPDIR

            export TMPDIR=/tmp/$USER;
            mkdir -p $TMPDIR;
            tensorboard --logdir ...

        code in line:
            export TMPDIR=/tmp/$USER; mkdir -p $TMPDIR; tensorboard --logdir ....

        solution is referred by https://github.com/tensorflow/tensorboard/issues/2010

        Returns:
            A SummaryWriter instance
        """
        from torch.utils.tensorboard import SummaryWriter

        kwargs = self.exp.board_args
        res = SummaryWriter(**kwargs)

        def close(*args):
            res.flush()
            res.close()

        self.exp.add_exit_hook(close)
        return res

    @property
    def datamodule(self):
        if self._datamodule is None:
            self._datamodule = DataModule()
        return self._datamodule

    @property
    def train_dataloader(self):
        return self.datamodule.train_dataloader

    @property
    def val_dataloader(self):
        return self.datamodule.val_dataloader

    @property
    def test_dataloader(self):
        return self.datamodule.test_dataloader

    def regist_device(self, device: torch.device):
        self._state_dicts[TRAINER.DKEY.device] = device
        if device.type == 'cuda':
            torch.cuda.set_device(device)

    def regist_dataloader(self, train: DataLoader = None, val: DataLoader = None, test: DataLoader = None,
                          datamodule: DataModule = None):
        if datamodule is not None:
            self._datamodule = datamodule
        else:
            self.datamodule.regist_dataloader(train=train, val=val, test=test)

    def add_callback(self, callback):
        """
        添加一个回调函数，注意，不能添加重复的 callback，这不推荐，也没有必要。
        :type callable,str
        :param callback:
        :return:
        """
        msg = None
        cb_name = callback.__class__.__name__
        if callback not in self._callback_set and cb_name in self._callback_name_set:
            msg = "Callback duplicate."
            callback.on_hook_failed(self, msg)
            return

        if callback.only_main_process and self.local_rank > 0:
            msg = f"{callback.__class__.__name__} only_main_process but in local_rank {self.local_rank}"
            callback.on_hook_failed(self, msg)
            return

        if callback.only_single_gpu and self.is_dist:
            msg = f"{callback.__class__.__name__} only_single_gpu but dist={self.is_dist}"
            callback.on_hook_failed(self, msg)
            return

        if msg is not None:
            return False
        bisect.insort(self._callback_set, callback)
        self._callback_name_set.add(cb_name)

        callback._hooked = self
        callback.on_hooked(self, self.params)
        self.logger.info("{} hooked on {}().".format(callback, self.__class__.__name__))
        return True

    def reload_callback(self, callback):
        """重新加载某 callback"""
        self.remove_callback(callback.__class__)
        return self.add_callback(callback)

    def remove_callback(self, callback):
        """
        移除已加载的 callback
        Args:
            callback: 可以是回调类名、实例、或回调类类型

        Returns:
            是否移除成功，若返回False，则表明没有找到对应的 callback
            若返回 True，则表明该 callback 已被完好移除
        """
        msg = None
        from .callbacks import BaseCallback

        try:
            if issubclass(callback, BaseCallback):
                for cb in self._callback_set:
                    if cb.__class__.__name__ == callback.__name__:
                        callback = cb
                        break
        except:  # handle TypeError: issubclass() arg 1 must be a class
            pass

        if isinstance(callback, str):
            for cb in self._callback_set:
                if cb.__class__.__name__ == callback:
                    callback = cb
                    break

        if callback not in self._callback_set:
            return False

        cb_name = callback.__class__.__name__
        self._callback_set.remove(callback)
        self._callback_name_set.remove(cb_name)
        self.logger.info("{} unhooked from {}.".format(callback, self))
        return True

    def change_mode(self, train=True):
        for k, v in self.model_dict.items():
            if train:
                v.train()
            else:
                v.eval()

    def optim_state_dict(self):
        return {k: v.state_dict() for k, v in self.optim_dict.items()}

    def model_state_dict(self):
        return {k: v.state_dict() for k, v in self.model_dict.items()}

    def buffer_state_dict(self):
        return self.buffers

    def other_state_dict(self):
        return {k: v.state_dict() for k, v in self.others.items()}

    def state_dict(self):
        res = {
            'models': self.model_state_dict(),
            'optims': self.optim_state_dict(),
            'other': self.other_state_dict(),
            'buffer': self.buffer_state_dict(),
        }
        for k, v in self._state_dicts.items():
            if k not in res:
                res[k] = v

        return res

    def _load_fun_state_dict(self, src: dict, tgt: dict):
        for k, v in tgt.items():
            if k in src:
                v.load_state_dict(src[k])

    def ignore_object_state_dict(self, item):
        pass

    def load_state_dict(self, object: Union[str, dict]):
        if isinstance(object, str):
            ckpt, meta_info = self.saver.load_state_dict(object, with_meta=True)
        else:
            ckpt, meta_info = object, None

        _sub = {'models', 'optims', 'other'}
        if ckpt is None:
            self.logger.error('state_dict object is None, `load_state_dict()` ignored.')
            return meta_info

        for k, v in ckpt.items():
            if k in _sub:
                self._load_fun_state_dict(v, self._state_dicts[k])
            else:
                self._state_dicts[k] = v
        return meta_info

    def _build_trainer_meta_info(self, meta_info: Union[str, dict, Meter] = None):
        info = Meter()
        info['eidx'] = self.eidx
        if meta_info is not None:
            if isinstance(meta_info, str):
                info['msg'] = meta_info
            if isinstance(meta_info, Meter):
                meta_info = meta_info.serialize()
            if isinstance(meta_info, dict):
                info.update(meta_info)
        return info.serialize()

    def save_checkpoint(self, max_keep=10, is_best=False, meta_info: Union[str, dict, Meter] = None):
        info = self._build_trainer_meta_info(meta_info)
        return self.saver.save_checkpoint(self.eidx, self.state_dict(),
                                          meta_info=info,
                                          max_keep=max_keep,
                                          is_best=is_best)

    def save_keypoint(self, meta_info: Union[str, dict, Meter] = None):
        info = self._build_trainer_meta_info(meta_info)
        return self.saver.save_keypoint(self.eidx, self.state_dict(),
                                        meta_info=info)

    def save_model(self, is_best=False, meta_info: Union[str, dict, Meter] = None):
        info = self._build_trainer_meta_info(meta_info)
        return self.saver.save_model(self.eidx, {'models': self.model_state_dict()},
                                     meta_info=info,
                                     is_best=is_best)


class DLLoopMix():
    def stop_train(self):
        raise NotImplementedError()

    def stop_train_epoch(self):
        raise NotImplementedError()

    def to_stage(self, stage: TrainerStage):
        raise NotImplementedError()

    def train(self, dataloader: Union[DataLoader, DataModule]):
        raise NotImplementedError()

    def train_epoch(self, dataloader: DataLoader):
        raise NotImplementedError()

    def train_step(self, idx, batch, *args, **kwargs):
        raise NotImplementedError()

    def evaluate(self, dataloader: Union[DataLoader, DataModule]):
        raise NotImplementedError()

    def evaluate_step(self, idx, batch, *args, **kwargs):
        raise NotImplementedError()

    def test(self, dataloader: Union[DataLoader, DataModule]):
        raise NotImplementedError()

    def test_step(self, idx, batch, *args, **kwargs):
        raise NotImplementedError()

    def inference(self, batch):
        raise NotImplementedError()

    def predict(self, batch):
        """alias of inference"""
        raise NotImplementedError()


class Trainer(DLLoopMix, _BaseTrainer):
    _call_backs = {
        "train", "train_epoch", "train_step",
        "test", "test_step",
        "evaluate", "evaluate_step",
        "predict", "inference",
        'prepare_dataloader',
    }

    def _wrap_result(self, meter: Union[Mapping, Meter, Sequence, torch.Tensor, np.ndarray]) -> dict:
        if meter is None:
            return {}
        if isinstance(meter, (Mapping, Meter)):
            return meter
        elif isinstance(meter, Sequence):
            return {f'm{i}': v for i, v in enumerate(meter)}
        elif isinstance(meter, (torch.Tensor, np.ndarray)):
            return {'metric': meter}

    @property
    def training(self):
        return self.params.stage == TrainerStage.train.name

    def stop_train(self):
        self.train_toggle = True

    def stop_train_epoch(self):
        self.train_epoch_toggle = True

    def to_stage(self, stage: TrainerStage):
        self.params.stage = stage.name
        self.change_mode(stage.name == TrainerStage.train.name)

    def to_device(self, item: Optional[Union[nn.Module, torch.Tensor, Sequence, Mapping]] = None,
                  device_args_kwargs=None):
        if device_args_kwargs is None:
            device_args_kwargs = self.device_arg_kwargs
        from lumo.utils.device import to_device
        if item is None:
            for k in self._models:
                v = self._models[k]
                if not self.accelerator.device_placement:
                    v.to(self.accelerator.device)
                v = self.accelerator.prepare(self._models[k])
                self._models[k] = v
            for k in self._optims:
                v = self.accelerator.prepare(self._optims[k])
                self._optims[k] = v
        else:
            return to_device(item, device_args_kwargs)

    def prepare_dataloader(self, stage: TrainerStage, dataloader=None) -> DataLoader:
        params = self.params
        initialized = getattr(self.initial, f"{stage.name}_dataloader")

        dataloader_ = None
        """use loader from passed params first, then use itself, finally"""
        if isinstance(dataloader, DataModuleMix):
            dataloader.idataloader(params, stage, initialized)
            self.regist_dataloader(datamodule=dataloader)
            dataloader_ = getattr(dataloader, f'{stage.name}_dataloader', None)
        elif isinstance(dataloader, DataLoader):
            self.regist_dataloader(**{stage.name: dataloader})
            dataloader_ = dataloader
        elif isinstance(self, DataModuleMix):
            self.idataloader(params, stage, initialized)
            dataloader_ = getattr(self, f'{stage.name}_dataloader', None)

        if dataloader is None and self.datamodule is not None:
            dataloader_ = getattr(self.datamodule, f'{stage.name}_dataloader', None)
            if dataloader_ is None:
                self.datamodule.idataloader(params, stage, initialized)
                dataloader_ = getattr(self.datamodule, f'{stage.name}_dataloader', None)

        if isinstance(dataloader_, DataLoader):
            dataloader_ = self.accelerator.prepare(dataloader_)

        kwargs = {stage.name: dataloader_}
        self.regist_dataloader(**kwargs)
        return dataloader_

    def train(self, dataloader: Union[DataLoader, DataModuleMix] = None) -> TrainerResult:
        # params, initialized = self.params, self.initial.train_dataloader
        dataloader = self.prepare_dataloader(TrainerStage.train, dataloader)
        if dataloader is None:
            return TrainerResult(TrainerStage.train, 1, 'no train_dataloader')

        self._check_models_init()
        self._check_optim_init()

        self.initial.train_dataloader = True

        result = TrainerResult(TrainerStage.train, None, TrainerResult.MSG_OK)
        while self.params.eidx < self.params.epoch:
            self.params.eidx += 1
            self.to_stage(TrainerStage.train)
            result = self.train_epoch(dataloader)
            if self.train_toggle:
                self.train_toggle = False
                return TrainerResult(TrainerStage.train, result.meter, TrainerResult.MSG_OK_MID)

            self.exp.dump_train_info(self.params.eidx)

        return TrainerResult(TrainerStage.train, result.meter, TrainerResult.MSG_OK)

    def train_epoch(self, dataloader: DataLoader) -> TrainerResult:
        avg = None
        for idx, batch in enumerate(dataloader):
            if avg is None:
                avg = AvgMeter()
            self.params.global_step += 1
            self.params.idx = idx
            meter = self.train_step(idx, batch, self.params)
            avg.update(self._wrap_result(meter))
            if self.train_epoch_toggle:
                self.train_epoch_toggle = False
                break
        return TrainerResult(TrainerStage.train_epoch, avg, TrainerResult.MSG_OK)

    def train_step(self, idx, batch, params: ParamsType, *args, **kwargs):
        pass

    def evaluate(self, dataloader: Union[DataLoader, DataModule] = None) -> TrainerResult:
        dataloader = self.prepare_dataloader(TrainerStage.val, dataloader)
        if dataloader is None:
            return TrainerResult(TrainerStage.val, None, TrainerResult.MSG_NO_DATALOADER)
        self._check_models_init()
        self.to_stage(TrainerStage.val)

        with torch.no_grad():
            avg = AvgMeter()
            for idx, batch in to_device_enumrate(dataloader, self.device_arg_kwargs):
                self.params.idx = idx
                meter = self.evaluate_step(idx, batch, self.params)
                avg.update(self._wrap_result(meter))

        return TrainerResult(TrainerStage.val, avg, TrainerResult.MSG_OK)

    def evaluate_step(self, idx, batch, params: ParamsType, *args, **kwargs) -> Meter:
        pass

    def test(self, dataloader: Union[DataLoader, DataModule] = None) -> TrainerResult:
        dataloader = self.prepare_dataloader(TrainerStage.test, dataloader)
        if dataloader is None:
            return TrainerResult(TrainerStage.test, None, TrainerResult.MSG_NO_DATALOADER)
        self._check_models_init()
        self.to_stage(TrainerStage.test)

        with torch.no_grad():
            avg = AvgMeter()
            for idx, batch in to_device_enumrate(dataloader, self.device_arg_kwargs):
                self.params.idx = idx
                meter = self.test_step(idx, batch, self.params)
                avg.update(self._wrap_result(meter))
        return TrainerResult(TrainerStage.test, avg, TrainerResult.MSG_OK)

    def test_step(self, idx, batch, params: ParamsType, *args, **kwargs):
        pass

    def inference(self, batch):
        pass

    def predict(self, batch):
        """alias of inference"""
        self.inference(batch)

    def ioptims(self, params: ParamsType):
        pass

    def icallbacks(self, params: ParamsType):
        pass

    def imodels(self, params: ParamsType):
        pass



class Compose:
    """Composes several transforms together. This transform does not support torchscript.
    Please, see the note below.

    Args:
        transforms (list of ``Transform`` objects): list of transforms to compose.

    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, mem: dict):
        for t in self.transforms:
            mem = t(mem)
        return mem

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        for t in self.transforms:
            format_string += '\n'
            format_string += '    {0}'.format(t)
        format_string += '\n)'
        return format_string


class WrapDict():
    def __init__(self, key, func):
        self.key = key
        self.func = func

    def __call__(self, *args, **kwargs):
        return {self.key: self.func(*args, **kwargs)}



"""

"""
import numpy as np
import torch

# from torchvision.transforms import *


class ToTensor(object):
    """Transform the image to tensor.
    """

    def __call__(self, x):
        x = torch.from_numpy(x)
        return x


class RandomVerticalFlip():
    def __init__(self, p=0.5):
        self.p = p

    def __call__(self, x):
        if np.random.rand() < self.p:
            x = np.flip(x, axis=0)

        return x


class ToNumpy():
    def __call__(self, x):
        return np.array(x)


class ToHWC():
    def __call__(self, x:np.ndarray):
        if x.ndim == 2:
            return x
        if x.shape[-1] <=4 :
            return x
        return x.transpose([2,0,1])

class ToCHW():
    def __call__(self, x:np.ndarray):
        if x.ndim == 2:
            return x
        if x.shape[-1] >4 :
            return x
        return x.transpose([2,0,1])

class RandomHorizontalFlip():
    """Flip randomly the image.
    """

    def __init__(self, p=0.5):
        self.p = p

    def __call__(self, x):
        if np.random.rand() < self.p:
            x = np.fliplr(x)
        return x.copy() # directly return x will cause  'negative' error when doing Torch.from_numpy

class RandomCrop():
    def __init__(self, size, padding=4, fill=0):
        self.size = size
        self.pad = Pad(pad=padding, fill=fill)

    def __call__(self, x):
        x = self.pad(x)

        if x.ndim == 2:
            h, w = x.shape
        elif x.shape[-1] <= 4:
            h, w = x.shape[:-1]
        else:
            h, w = x.shape[1:]

        new_h = new_w = self.size

        top = np.random.randint(0, h - new_h)
        left = np.random.randint(0, w - new_w)

        if x.ndim == 2:
            x = x[top: top + new_h, left: left + new_w]
        elif x.shape[-1] <= 4:
            x = x[top: top + new_h, left: left + new_w, :]
        else:
            x = x[:, top: top + new_h, left: left + new_w]

        return x


class Pad():
    def __init__(self, pad=4, fill=0):
        self.pad = pad
        self.fill = fill

    def __call__(self, x):
        if x.ndim == 2:
            x = np.pad(x, self.pad, mode="constant")
        else:
            if x.shape[-1] <= 4:
                x = np.pad(x, ((self.pad, self.pad), (self.pad, self.pad), (0, 0)), mode="constant",
                           constant_values=self.fill)
            else:
                x = np.pad(x, ((0, 0), (self.pad, self.pad), (self.pad, self.pad)), mode="constant",
                           constant_values=self.fill)
        return x



def copy_templete():
    import os
    import shutil

    src = 'lumo-templete'
    dst = 'lumo/cli/templete'
    if os.path.exists(dst):
        shutil.rmtree(dst)

    shutil.copytree(src, dst)

    for root, dirs, fs in os.walk(dst):
        for f in fs:
            if f.endswith('.py'):
                f = os.path.join(root, f)
                prefn, ext = os.path.splitext(f)

                os.rename(f, '{}.py-tpl'.format(prefn))



import os
from lumo.base_classes.tree import tree
from typing import Callable, Union


def touch(file, content=None):
    if content is None:
        content = ''
    f = open(file, 'w', encoding='utf8')
    f.write(content)
    f.close()


class Template:
    def __init__(self):
        self.tree = tree()

    def add_directory(self, path, *paths):
        paths = [path] + list(paths)
        tree = self.tree
        for path in paths:
            for sub in path.split('/'):
                tree = tree[sub]
        return self

    def add_file(self, path, *paths, info: Union[Callable, str] = None):
        paths = [path] + list(paths)
        pre = self.tree
        tree = self.tree
        for path in paths:
            for sub in path.split('/'):
                pre = tree
                tree = tree[sub]
        pre[sub] = info
        return self

    def create(self, root='./', ignore_if_not_empty=True):
        if os.path.exists(root) and len(os.listdir(root)) and ignore_if_not_empty:
            return False

        for file, info in self.tree.walk():
            file = os.path.join(root, file)
            os.makedirs(os.path.dirname(file), exist_ok=True)
            if callable(info):
                info(file)
            elif isinstance(info, dict):
                os.makedirs(file, exist_ok=True)
            else:
                touch(file, info)
        return True



import os
from collections import namedtuple
from lumo.base_classes.db import _NoneResult, MPStruct

row = namedtuple('row', ['id', 'value'])


class Queue(MPStruct):
    TABLE_NAME = 'QUEUE'
    TABLE_SQL = """
        CREATE TABLE IF NOT EXISTS QUEUE
           (ID      INTEGER PRIMARY KEY autoincrement,
           VALUE    TEXT    NOT NULL);
        """

    INDEX_SQL = """
           CREATE UNIQUE INDEX QVALUE_INDEX on QUEUE (VALUE)
    """

    def init_table(self):
        if self.has_table(self.TABLE_NAME):
            return

        conn = self.connect
        c = conn.cursor()
        c.executescript('\n'.join([self.TABLE_SQL]))

        try:
            c.execute(self.INDEX_SQL)
        except:
            pass
        conn.commit()

    def _del_rec(self, id, table):
        res = self.execute(f"delete from {table} where id={id};", 'w')
        return not isinstance(res, _NoneResult) and res.rowcount > 0

    def _del_recs(self, *ids, table):
        ids = ','.join([str(i) for i in ids])
        res = self.execute(f"delete from {table} where id in ({ids});", 'w')
        return not isinstance(res, _NoneResult) and res.rowcount > 0

    def push(self, value):
        if self.value_in_queue(value) is not None:
            return False
        value = self.encode_value(value)

        res = self.execute(f"insert into queue (value) values ('{value}');", 'w')
        return not isinstance(res, _NoneResult)

    def pushk(self, *values):
        values = [self.encode_value(i) for i in values if self.value_in_queue(i) is None]

        if len(values) == 0:
            return 0

        inner = ','.join([f"('{value}')" for value in values])

        res = self.execute(f"insert into queue (value) values {inner};", 'w')
        if isinstance(res, _NoneResult):
            return -1
        else:
            return len(values)

    def popk(self, k=1):
        rec = self.execute(f"select id,value from queue limit {k};").fetchall()

        if len(rec) > 0:
            ids = [i[0] for i in rec]
            values = [self.decode_value(i[1]) for i in rec]

            if not self._del_recs(*ids, table=self.TABLE_NAME):
                return self.popk(k)
            return [row(id, val) for id, val in zip(ids, values)]
        return []

    def pop(self):
        rec = self.execute(f"select id,value from queue limit 1;").fetchone()
        if rec is not None:
            id, value = rec
            value = self.decode_value(value)
            if not self._del_rec(id, self.TABLE_NAME):
                return self.pop()
            return row(id, value)
        return None

    def value_in_queue(self, value):
        value = self.encode_value(value)
        res = self.execute(f"select id from queue where value='{value}';").fetchone()
        if res is not None:
            return res[0]
        else:
            return None

    def top(self):
        rec = self.execute(f"select id,value from queue limit 1;").fetchone()
        if rec is not None:
            id, value = rec
            value = self.decode_value(value)
            return row(id, value)
        return None

    @property
    def count(self):
        return self.execute('select count(id) from queue').fetchone()[0]


class Bucket(Queue):

    def __init__(self, session_id=None, root=None, retry=7):
        super().__init__(session_id, root, retry)
        self.file = os.path.join(self.root, f"{Queue.__name__}_{session_id}.sqlite")

    def bucket(self, ind, total):
        rec = self.execute(f"select id,value from queue where id % {total}={ind}").fetchall()
        ids = [i[0] for i in rec]
        values = [self.decode_value(i[1]) for i in rec]
        return [row(id, val) for id, val in zip(ids, values)]

    def iter_bucket(self, ind, total):
        for rec in self.execute(f"select id,value from queue where id % {total}={ind}"):
            yield row(rec[0], self.decode_value(rec[1]))


class Marker(MPStruct):

    def mark(self, key, tag):
        pass

    def unmark(self, key, tag):
        pass

    def has_marks(self, key, *tag):
        pass

    def get_marks(self, key):
        pass


class Dict(MPStruct):
    TABLE_NAME = 'DICT'
    TABLE_SQL = '''
    CREATE TABLE IF NOT EXISTS DICT
       (ID      INTEGER PRIMARY KEY autoincrement,
       KEY      TEXT    NOT NULL,
       VALUE    TEXT    NOT NULL);
    CREATE UNIQUE INDEX KEY_INDEX on DICT (KEY);
       '''

    def init_table(self):
        if self.has_table(self.TABLE_NAME):
            return

        conn = self.connect
        c = conn.cursor()
        c.executescript('\n'.join([self.TABLE_SQL]))
        conn.commit()

    def set(self, key, value: str):
        res = self.get(key)
        value = self.encode_value(value)

        if res is None:
            self.execute(f"insert into dict (key,value) values ('{key}','{value}');")

    def get(self, key, default=None):
        res = self.execute(f"select id,value from dict where key='{key}';").fetchone()
        if res is None:
            return default
        else:
            return row(res[0], self.decode_value(res[1]))


if __name__ == '__main__':
    print()



# code in this file is adpated from
# https://github.com/ildoonet/pytorch-randaugment/blob/master/RandAugment/augmentations.py
# https://github.com/google-research/fixmatch/blob/master/third_party/auto_augment/augmentations.py
# https://github.com/google-research/fixmatch/blob/master/libml/ctaugment.py
import random

import numpy as np
import PIL
import PIL.ImageOps
import PIL.ImageEnhance
import PIL.ImageDraw
from PIL import Image

PARAMETER_MAX = 10


def AutoContrast(img, **kwarg):
    return PIL.ImageOps.autocontrast(img)


def Brightness(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    return PIL.ImageEnhance.Brightness(img).enhance(v)


def Color(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    return PIL.ImageEnhance.Color(img).enhance(v)


def Contrast(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    return PIL.ImageEnhance.Contrast(img).enhance(v)


def Cutout(img, v, max_v, bias=0):
    if v == 0:
        return img
    v = _float_parameter(v, max_v) + bias
    v = int(v * min(img.size))
    return CutoutAbs(img, v)


def CutoutAbs(img, v, **kwarg):
    w, h = img.size
    x0 = np.random.uniform(0, w)
    y0 = np.random.uniform(0, h)
    x0 = int(max(0, x0 - v / 2.))
    y0 = int(max(0, y0 - v / 2.))
    x1 = int(min(w, x0 + v))
    y1 = int(min(h, y0 + v))
    xy = (x0, y0, x1, y1)
    # gray
    color = (127, 127, 127)
    img = img.copy()
    PIL.ImageDraw.Draw(img).rectangle(xy, color)
    return img


def Equalize(img, **kwarg):
    return PIL.ImageOps.equalize(img)


def Identity(img, **kwarg):
    return img


def Invert(img, **kwarg):
    return PIL.ImageOps.invert(img)


def Posterize(img, v, max_v, bias=0):
    v = _int_parameter(v, max_v) + bias
    return PIL.ImageOps.posterize(img, v)


def Rotate(img, v, max_v, bias=0):
    v = _int_parameter(v, max_v) + bias
    if random.random() < 0.5:
        v = -v
    return img.rotate(v)


def Sharpness(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    return PIL.ImageEnhance.Sharpness(img).enhance(v)


def ShearX(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    if random.random() < 0.5:
        v = -v
    return img.transform(img.size, PIL.Image.AFFINE, (1, v, 0, 0, 1, 0))


def ShearY(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    if random.random() < 0.5:
        v = -v
    return img.transform(img.size, PIL.Image.AFFINE, (1, 0, 0, v, 1, 0))


def Solarize(img, v, max_v, bias=0):
    v = _int_parameter(v, max_v) + bias
    return PIL.ImageOps.solarize(img, 256 - v)


def SolarizeAdd(img, v, max_v, bias=0, threshold=128):
    v = _int_parameter(v, max_v) + bias
    if random.random() < 0.5:
        v = -v
    img_np = np.array(img).astype(np.int)
    img_np = img_np + v
    img_np = np.clip(img_np, 0, 255)
    img_np = img_np.astype(np.uint8)
    img = Image.fromarray(img_np)
    return PIL.ImageOps.solarize(img, threshold)


def TranslateX(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    if random.random() < 0.5:
        v = -v
    v = int(v * img.size[0])
    return img.transform(img.size, PIL.Image.AFFINE, (1, 0, v, 0, 1, 0))


def TranslateY(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    if random.random() < 0.5:
        v = -v
    v = int(v * img.size[1])
    return img.transform(img.size, PIL.Image.AFFINE, (1, 0, 0, 0, 1, v))


def _float_parameter(v, max_v):
    return float(v) * max_v / PARAMETER_MAX


def _int_parameter(v, max_v):
    return int(v * max_v / PARAMETER_MAX)


def fixmatch_augment_pool():
    # FixMatch paper
    augs = [(AutoContrast, None, None),
            (Brightness, 0.9, 0.05),
            (Color, 0.9, 0.05),
            (Contrast, 0.9, 0.05),
            (Equalize, None, None),
            (Identity, None, None),
            (Posterize, 4, 4),
            (Rotate, 30, 0),
            (Sharpness, 0.9, 0.05),
            (ShearX, 0.3, 0),
            (ShearY, 0.3, 0),
            (Solarize, 256, 0),
            (TranslateX, 0.3, 0),
            (TranslateY, 0.3, 0)]
    return augs


def my_augment_pool():
    # Test
    augs = [(AutoContrast, None, None),
            (Brightness, 1.8, 0.1),
            (Color, 1.8, 0.1),
            (Contrast, 1.8, 0.1),
            (Cutout, 0.2, 0),
            (Equalize, None, None),
            (Invert, None, None),
            (Posterize, 4, 4),
            (Rotate, 30, 0),
            (Sharpness, 1.8, 0.1),
            (ShearX, 0.3, 0),
            (ShearY, 0.3, 0),
            (Solarize, 256, 0),
            (SolarizeAdd, 110, 0),
            (TranslateX, 0.45, 0),
            (TranslateY, 0.45, 0)]
    return augs


class RandAugmentPC(object):
    def __init__(self, n, m):
        assert n >= 1
        assert 1 <= m <= 10
        self.n = n
        self.m = m
        self.augment_pool = my_augment_pool()

    def __call__(self, img):
        ops = random.sample(self.augment_pool, k=self.n)
        for op, max_v, bias in ops:
            prob = np.random.uniform(0.2, 0.8)
            if random.random() + prob >= 1:
                img = op(img, v=self.m, max_v=max_v, bias=bias)
        img = CutoutAbs(img, 16)
        return img


class RandAugmentMC(object):
    """输入为 PIL.Image，输出为增广后的 PIL.Image"""

    def __init__(self, n, m):
        assert n >= 1
        assert 1 <= m <= 10
        self.n = n
        self.m = m
        self.augment_pool = fixmatch_augment_pool()

    def __call__(self, img):
        ops = random.sample(self.augment_pool, k=self.n)
        for op, max_v, bias in ops:
            v = np.random.randint(1, self.m)
            if random.random() < 0.5:
                img = op(img, v=v, max_v=max_v, bias=bias)
        img = CutoutAbs(img, 16)
        return img



"""

"""


# from torch.optim.adadelta import Adadelta
# from torch.optim.adagrad import Adagrad
# from torch.optim.adam import Adam
# from torch.optim.adamax import Adamax
# from torch.optim.adamw import AdamW
# from torch.optim.asgd import ASGD
# from torch.optim.lbfgs import LBFGS
# from torch.optim.optimizer import Optimizer
# from torch.optim.rmsprop import RMSprop
# from torch.optim.rprop import Rprop
# from torch.optim.sgd import SGD
# from torch.optim.sparse_adam import SparseAdam
# from torch.optim import adam,adamw,adamax,adagrad,adadelta,asgd,sparse_adam,sgd,lr_scheduler,lbfgs,optimizer,rmsprop,rprop



import os
import sys
from sys import platform
from typing import Mapping, TypeVar, Type

T = TypeVar('T')

from lumo import __version__
from lumo.contrib.itertools import lfilter

LIBRARY_NAME = 'lumo'


class ENV:
    IS_WIN = (platform == "win32")
    IS_MAC = (platform == "darwin")
    IS_LINUX = (platform == "linux" or platform == "linux2")

    IS_REMOTE = any([i in os.environ for i in ['SHELL',
                                               'SHLVL',
                                               'SSH_CLIENT',
                                               'SSH_CONNECTION',
                                               'SSH_TTY']])
    IS_LOCAL = not IS_REMOTE

    IS_PYCHARM = os.environ.get("PYCHARM_HOSTED", '0') == "1"
    IS_PYCHARM_DEBUG = eval(os.environ.get('IPYTHONENABLE', "False"))

    ENV = "jupyter" if "jupyter_core" in sys.modules else "python"

    IS_GIT_ENABLED = (os.environ.get('LUMO_GIT', '1') == '1' and
                      os.environ.get('LUMO_NOGIT', '0') == '0')


class CFG:
    class PATH:  # key for path
        GLOBAL_EXP = 'global_exp'
        DATASET = 'datasets'
        PRETRAINS = 'pretrains'
        CACHE = 'cache'

    class FIELD:
        RUNTIME = 'runtime'
        GLOBAL = 'global'
        DEFAULT = 'default'

    BRANCH_NAME = 'experiment'


class FN:
    PHASH = '.hash'
    # CONFIGJS = 'config.json'
    REPOSJS = 'repos.json'
    TESTLOG = 'tests.log'
    VERSION = f'.lumo.{__version__}'

    D_LINE = '.line'
    D_JSON = '.json'
    D_PKL = '.pkl'

    class SUFFIX:
        D_LINE = '.txt'
        D_JSON = '.json'
        D_PKL = '.pkl'


class EXP:  # used in Experiment
    STATE = 'state'  # dumped info key
    EXCEPTION = 'exception'
    PROJECT = 'project'
    EXECUTE = 'execute'
    GIT = 'git'

    VERSION = 'version'
    UUID = 'uuid'

    TRAINER = 'trainer'


class TRAINER:
    path = 'path'
    doc = 'doc'
    basename = 'basename'
    class_name = 'class_name'

    class DKEY:
        models = 'models'
        optims = 'optims'
        tensor = 'tensor'
        others = 'others'
        initial = 'initial'
        train_epoch_toggle = 'train_epoch_toggle'
        train_toggle = 'train_toggle'
        device = 'device'
        devices = 'devices'
        ini_models = 'initial.models'
        ini_datasets = 'initial.datasets'
        ini_callbacks = 'initial.callbacks'
        params = 'params'
        exp = 'exp'


class EXP_CONST:
    class INFO_KEY:
        STATE = 'state'  # dumped info key
        PROJECT = 'project'
        EXECUTE = 'execute'
        GIT = 'git'
        VERSION = 'version'
        TRAINER = 'trainer'

    class IO_DIR:
        SINFO_DIR = '.json'
        PKL_DIR = '.pkl'
        INFO_DIR = '.line'

    # class HOOK_FN


"""

"""

from lumo import Params
params = Params()
params.from_args()
print(params)

params.epoch = 400
params.batch_size = 25
print(params)


from lumo import Params
class MyParams(Params):
    def __init__(self):
        super().__init__()
        self.batch_size = 50
        self.topk = (1,2,3,4)
        self.optim = dict(
            lr=0.009,
            moment=0.9
        )

params = MyParams()
print(params)


from lumo import Params
params = Params()
params.choice("dataset","mnist","cifar10","cifar100","svhn")
params.arange("thresh",5,0,20)
print(params)

# for g in params.grid_search("thresh",range(0,20)):
#     for g in g.grid_search("dataset",['cifar10','cifar100','svhn']):
#         print(g.dataset,g.thresh)


params.bind('dataset','mnist','arch','simplenet')
params.bind('dataset','cifar10','arch','cnn13')
params.bind('arch','simplenet','arch_param',dict(feature=128))
params.bind('arch','cnn13','arch_param',dict(feature=256))
params.dataset = 'cifar10'
print(params.arch)
print(params.arch_param)
params.dataset = 'mnist'
print(params.arch)
print(params.arch_param)



params.to_json('params.json')
params.from_json('params.json')


p = Params()
p.margin = 0.5
p.margin = p.default(0.1,True)


p = Params()
p.margin = p.default(0.3,True)
print(p.margin)

p.optim = p.create_optim('SGD',lr=0.1)
from torch import nn
p.optim.build(nn.Linear(2,2).parameters())
cp = p._param_dict.copy()



print(cp)




from . import builder
from . import exphook

from .builder import DatasetBuilder
from .delegate import DataDelegate

from .environ import globs
from .experiment import Experiment, TrainerExperiment
from .exphook import ExpHook
from .finder import F, Q

from .logger import Logger
from .meter import Meter, AvgMeter
from .params import Params, BaseParams, ParamsType, DistributionParams

from .random import Random
from .saver import Saver

from .trainer import Trainer
from .mixin import ModelMix, CallbackMix, DataModuleMix
from . import callbacks

from .datamodule import DataModule



"""

"""
import sys
sys.path.insert(0,"../")
from lumo import __version__
import time
print(__version__)

import random
# from lumo.utils import random as rnd
import torch.nn as nn
import torch
from torch.optim import SGD


from lumo import RndManager
rnd = RndManager()

for i in range(2):
    rnd.mark("train")
    # rnd.fix_seed(1)
    data = torch.rand(5, 2)
    y = torch.tensor([0, 0, 0, 0, 0])
    model = nn.Linear(2, 2)

    sgd = SGD(model.parameters(), lr=0.01)
    logits = model(data)
    loss = nn.CrossEntropyLoss()(logits, y)
    loss.backward()
    sgd.step()
    sgd.zero_grad()

    print(list(model.parameters()))

    rnd.shuffle()
    print(random.random())



"""

"""
import sys
sys.path.insert(0,"../")
from lumo import __version__
print(__version__)

from lumo import Experiment,globs

exp = Experiment("expname")

globs['a'] = 4

from pprint import pprint
pprint(globs.items())
pprint(exp.config_items())


"""

"""

import torch

from lumo.calculate import accuracy as acc


def test_classify():
    labels = torch.tensor([0, 1, 2, 3])
    preds = torch.tensor([[5, 4, 3, 2], [5, 4, 3, 2], [5, 4, 3, 2], [5, 4, 3, 2]], dtype=torch.float)
    total, res = acc.classify(preds, labels, topk=(1, 2, 3, 4))
    assert total == 4
    assert res[0] == 1 and res[1] == 2 and res[2] == 3 and res[3] == 4,str(res)




def find_free_network_port() -> int:
    """
    Finds a free port on localhost.
    It is useful in single-node training when we don't want to connect to a real master node but
    have to set the `MASTER_PORT` environment variable.
    """
    import socket
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port


"""

"""
from functools import wraps
from typing import Any, Mapping, Sequence

import numpy as np
import torch

from torch.utils.data._utils.collate import default_collate


class CollateBase():

    def __new__(cls, *args, **kwargs) -> Any:
        self = super().__new__(cls)

        def wrap(func):
            @wraps(func)
            def inner(*args, **kwargs):
                res = self.before_collate(*args, **kwargs)
                res = func(res)
                res = self.after_collate(res)
                return res

            return inner

        self.collate = wrap(self.collate)
        return self

    def __init__(self, collate_fn=default_collate, *args, **kwargs) -> None:
        super().__init__()
        self._collate_fn = collate_fn
        self.initial(*args, **kwargs)

    def initial(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return self.collate(*args, **kwargs)

    def before_collate(self, sample_list):
        return sample_list

    def collate(self, sample_list):
        return self._collate_fn(sample_list)

    def after_collate(self, batch):
        return batch


class IgnoreNoneCollate(CollateBase):

    def _filter_none(self, item):
        if item is None:
            return False
        if isinstance(item, (list, tuple)):
            return all([self._filter_none(i) for i in item])
        if isinstance(item, dict):
            return all(self._filter_none(i) for i in item.values())
        return True

    def before_collate(self, sample_list):
        return list(filter(self._filter_none, sample_list))


def numpy_collate(batch):
    r"""Puts each data field into a tensor with outer dimension batch size"""

    elem = batch[0]
    elem_type = type(elem)
    if isinstance(elem, torch.Tensor):
        batch = [elem.detach().cpu().numpy() for elem in batch]
        return np.stack(batch, 0)
    elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' \
            and elem_type.__name__ != 'string_':
        if elem_type.__name__ == 'ndarray' or elem_type.__name__ == 'memmap':
            return numpy_collate([np.array(b) for b in batch])
        elif elem.shape == ():  # scalars
            return np.array(batch)
    elif isinstance(elem, float):
        return np.array(batch, dtype=np.float)
    elif isinstance(elem, int):
        return np.array(batch)
    elif isinstance(elem, (str, bytes)):
        return batch
    elif isinstance(elem, Mapping):
        return {key: numpy_collate([d[key] for d in batch]) for key in elem}
    elif isinstance(elem, tuple) and hasattr(elem, '_fields'):  # namedtuple
        return elem_type(*(numpy_collate(samples) for samples in zip(*batch)))
    elif isinstance(elem, Sequence):
        # check to make sure that the elements in batch have consistent size
        it = iter(batch)
        elem_size = len(next(it))
        if not all(len(elem) == elem_size for elem in it):
            raise RuntimeError('each element in list of batch should be of equal size')
        transposed = zip(*batch)
        return [numpy_collate(samples) for samples in transposed]



import sys
from functools import wraps


def replace(func):
    sys.excepthook = func


def wrap_after(func):
    old = sys.excepthook

    def outer(fun):
        @wraps(fun)
        def inner(*args, **kwargs):
            old(*args, **kwargs)
            fun(*args, **kwargs)

        return inner

    sys.excepthook = outer(func)


def wrap_before(func):
    old = sys.excepthook

    def outer(fun):
        @wraps(fun)
        def inner(*args, **kwargs):
            fun(*args, **kwargs)
            old(*args, **kwargs)

        return inner

    sys.excepthook = outer(func)



"""

"""
import copy
import warnings
from collections import OrderedDict
from typing import Any, Iterable

import numpy as np
import torch

from lumo.base_classes.metaclasses import meta_attr

# from lumo.utils.keys import
_attr_clss = {}
from lumo.utils.hash import hash

ATTR_TYPE = '_type'


class attr(OrderedDict, metaclass=meta_attr):
    """
    An ordered defaultdict, the default class is attr itself.
    """

    def __new__(cls, *args: Any, **kwargs: Any):
        self = super().__new__(cls, *args, **kwargs)
        # self._class_name = cls.__name__
        return self

    @staticmethod
    def __parse_value(v):
        if isinstance(v, attr):
            pass
        elif isinstance(v, dict):
            v = attr.from_dict(v)
        return v

    def __getattr__(self, item):
        if item not in self:
            self[item] = attr()
        return self[item]

    def __setattr__(self, name: str, value: Any) -> None:
        value = self.__parse_value(value)
        self[name] = value

    def __getitem__(self, k):
        cwk = k
        if isinstance(k, _contains):
            k = k.value

        k = str(k)
        ks = k.split(".")
        if len(ks) == 1:
            if isinstance(cwk, _contains):
                return super().__getitem__(ks[0])

            try:
                return super().__getitem__(ks[0])
            except KeyError:
                self[ks[0]] = attr()
                return self[ks[0]]

        cur = self
        for tk in ks:
            if isinstance(cwk, _contains):
                cur = cur.__getitem__(_contains(tk))
            else:
                cur = cur.__getitem__(tk)
        return cur

    def __setitem__(self, k, v):
        v = self.__parse_value(v)

        k = str(k)
        ks = k.split(".")
        if len(ks) == 1:
            super().__setitem__(ks[0], v)
        else:
            cur = self
            for tk in ks[:-1]:
                ncur = cur.__getattr__(tk)
                if not isinstance(ncur, attr):
                    cur[tk] = attr()
                    ncur = cur[tk]
                cur = ncur

            cur[ks[-1]] = v

    def __contains__(self, o: object) -> bool:
        try:
            _ = self[_contains(o)]
            return True
        except:
            return False

    def __copy__(self):
        # res = attr()
        return self.from_dict(self)

    def __deepcopy__(self, memodict={}):
        return self.copy()

    def walk(self):
        for k, v in self.items():
            if isinstance(v, attr):
                for ik, iv in v.walk():
                    ok = "{}.{}".format(k, ik)
                    yield ok, iv
            else:
                yield k, v

    def __eq__(self, o: object) -> bool:
        if isinstance(o, dict):
            return self.hash() == hash(o)
        return False

    def items(self, toggle=False):
        if toggle:
            yield ATTR_TYPE, self.__class__.__name__
        for k, v in super().items():
            yield k, v

    def raw_items(self):
        return self.items(toggle=True)

    def pickify(self) -> dict:
        """
        Return a serialized dict which can be dumped by pickle.
        You can deserialize this dict by `from_dict()`
        """
        res = dict()
        for k, v in self.raw_items():
            if isinstance(v, attr):
                v = v.pickify()
                res[k] = v
            elif isinstance(v, (Iterable)):
                if not isinstance(v, (torch.Tensor, np.ndarray, str)):
                    nv = []
                    for vv in v:
                        if isinstance(vv, dict):
                            if isinstance(vv, attr):
                                nv.append(vv.pickify())
                            else:
                                nv.append(attr.from_dict(vv).pickify())
                        elif vv is None:
                            nv.append(_none().pickify())
                        else:  # set,list
                            # _tmp = attr()
                            # _tmp['tmp'] = vv
                            nv.append(vv)
                else:
                    nv = v
                res[k] = nv
            elif v is None:
                res[k] = _none().pickify()
            else:
                res[k] = v

        return res

    def jsonify(self) -> dict:
        """
        Return a serialized dict which can be dumped in json format.
        You can deserialize this dict by `from_dict()`
        """
        import numbers
        res = dict()
        for k, v in self.raw_items():
            if isinstance(v, (numbers.Number, str)):
                res[k] = v
            elif isinstance(v, attr):
                v = v.jsonify()
                res[k] = v
            elif isinstance(v, (Iterable)):
                if isinstance(v, (torch.Tensor, np.ndarray)):
                    nv = _tpttr(v).jsonify()
                else:
                    nv = []
                    for vv in v:
                        if isinstance(vv, (numbers.Number, str)):
                            nv.append(vv)
                        elif isinstance(vv, dict):
                            if isinstance(vv, attr):
                                nv.append(vv.jsonify())
                            else:
                                nv.append(attr.from_dict(vv).jsonify())
                        elif vv is None:
                            nv.append(_none().jsonify())
                        else:  # set,list
                            _tmp = attr()
                            _tmp['tmp'] = vv
                            nv.append(_tmp.jsonify()['tmp'])
                res[k] = nv
            elif v is None:
                res[k] = _none().jsonify()

        return res

    def hash(self) -> str:
        """return a hash string, both order/key/value changes will change the hash value."""
        return hash(self)

    def copy(self):
        """deep copy"""
        return self.__copy__()

    def replace(self, **kwargs):
        """An alias of update()"""
        for k, v in kwargs.items():
            self[k] = v
        return self

    @classmethod
    def from_dict(cls, dic: dict):
        # res = cls()
        cls_name = dic.get(ATTR_TYPE, cls.__name__)
        if cls_name not in _attr_clss:
            res = attr()
            from .errors import AttrTypeNotFoundWarning
            warnings.warn('{} not found, will use class attr to receive values.'.format(cls_name),
                          AttrTypeNotFoundWarning)
        else:
            res = _attr_clss[cls_name].__new__(_attr_clss[cls_name])

        if ATTR_TYPE in dic:
            dic.pop(ATTR_TYPE)

        if isinstance(res, _none):
            return None
        elif isinstance(res, _tpttr):
            if dic['torch']:
                return torch.tensor(dic['arr'])
            else:
                return np.array(dic['arr'])

        for k, v in dic.items():
            if isinstance(v, dict):
                v = attr.from_dict(v)
            elif isinstance(v, (list, tuple, torch.Tensor, np.ndarray)):
                v = cls._copy_iters(v)

            res[k] = v
        return res

    @staticmethod
    def _copy_iters(item):
        if isinstance(item, torch.Tensor):
            return item.clone()
        elif isinstance(item, np.ndarray):
            return item.copy()
        elif isinstance(item, list):
            return list(attr._copy_iters(i) for i in item)
        elif isinstance(item, set):
            return set(attr._copy_iters(i) for i in item)
        elif isinstance(item, tuple):
            return tuple(attr._copy_iters(i) for i in item)
        elif isinstance(item, dict):
            return attr.from_dict(item)
        return copy.copy(item)


class _none(attr):
    """
    A tricky item used in attr to replace None value when serializing.
    """


class _tpttr(attr):
    """
    A tricky item used in attr to parse pytorch and numpy array.
    """

    def __init__(self, arr):
        super().__init__()
        self.arr = arr.tolist()
        self.torch = isinstance(arr, torch.Tensor)


class _contains():
    """
    A tricky item that used in attr to discriminate the difference between `get()` and `contains()`
    """

    def __init__(self, value):
        self.value = value



"""
Methods about files/paths/hash
"""
import os


def checkpath(*path):
    res = os.path.join(*path)
    os.makedirs(res, exist_ok=True)
    if not os.path.exists(res):
        return None
    return res


def compare_path(a, b):
    if a is None and b is None:
        return True
    if a is None or b is None:
        return False

    a, b = os.path.realpath(a), os.path.realpath(b)
    return os.path.normpath(a) == os.path.normpath(b)


def walk_files(root, suf='.ft'):
    if suf != '':
        suf = f'{suf.lstrip(".")}'  # check suffix
    for root, dirs, fs in os.walk(root):
        for f in sorted(fs):
            if f.endswith(suf):
                yield os.path.join(root, f)



import os
import shutil
from collections import OrderedDict
from copy import copy
from itertools import accumulate
from operator import add
from typing import Sized, Sequence, Union, Callable, Dict, Any, List, overload, TypeVar, ClassVar
from functools import wraps
from torch.utils.data import Dataset, sampler as sp, DataLoader
import numpy as np
from lumo.contrib.data.collate import CollateBase
from .delegate import DataDelegate, Data, DelegateDataTypeError

Loader = TypeVar('Loader')


def _loader_partial(loader: Loader, *args, **kwargs) -> Loader:
    args = list(args)

    @wraps(loader)
    def DataLoader(*_, **ikwargs):
        nonlocal args
        if len(_) > 0:
            raise ValueError('lumo makes DataLoader takes 0 positional arguments, please pass all arguments with key')
        kwargs.update(ikwargs)

        return loader(*args, **kwargs)

    return DataLoader


class BaseBuilder(Dataset):

    def __init__(self):
        self._check_delegate_flag = set()

        self._sampler = None
        self._batch_sampler = None
        self._chain = False
        self._sub_indices = None
        self._reindices = None
        self._dataloader_kwargs = None

    def __repr__(self):
        return (f"{self.__class__.__name__}()")

    def __getitem__(self, index):
        raise NotImplementedError()

    def __len__(self):
        if self._sub_indices is not None:
            return len(self._sub_indices)
        return self.raw_len

    @property
    def sampler(self):
        return self._sampler

    @property
    def batch_sampler(self):
        return self._batch_sampler

    @property
    def raw_len(self):
        raise NotImplementedError()

    def _check_len(self, source: Sized):
        """
        Each source must have equal size, this function will check this, and an Exception will be raised
        if new source doesn't have the same size that added sources have.
        """
        len_ = len(source)
        if self._dataset_len is None:
            self._dataset_len = len_
        else:
            assert self._dataset_len == len(source), f"{self._dataset_len} != {len(source)}"

    def _map_index(self, index):
        """
        Map the raw index to the final index for source data.
        Args:
            index:

        Returns:

        """
        if self._sub_indices is not None:
            index = self._sub_indices[index]
        if self._reindices is not None:
            index = self._reindices[index]

        return index

    def copy(self):
        builder = DatasetBuilder()
        builder._sub_indices = copy(self._sub_indices)
        builder._reindices = copy(self._reindices)
        return builder

    def subset(self, indices: Sequence[int], copy=False):
        indices = np.array(indices)
        if copy:
            builder = self.copy()
            builder.subset(indices, copy=False)
            return builder

        self._sub_indices = indices
        return self

    def reindices(self, indices: Sequence[int]):
        self._reindices = np.array(indices)
        return self

    def zip(self):
        self._chain = False
        return self

    def chain(self):
        self._chain = True
        return self

    def shuffle(self, flag=True):
        """
        shuffle before dataloader may decrease memory leakage
            see https://github.com/pytorch/pytorch/issues/13246#issuecomment-436632186 for details
        Returns:

        """
        if not flag:
            return self
        if self._sub_indices is not None:
            np.random.shuffle(self._sub_indices)
        else:
            if self._reindices is None:
                self.reindices(range(len(self)))
            np.random.shuffle(self._reindices)
        return self

    def random_sampler(self, replacement: bool = False, generator=None):
        _sampler = sp.RandomSampler(self, replacement=replacement, generator=generator)
        self.sample_by(_sampler)
        return self

    def weighted_random_sampler(self, weights: Sequence[float],
                                replacement: bool = True, generator=None):
        _sampler = sp.WeightedRandomSampler(weights=weights,
                                            num_samples=len(self),
                                            replacement=replacement,
                                            generator=generator)
        self.sample_by(_sampler)
        return self

    def sequential_sampler(self):
        _sampler = sp.SequentialSampler(self)
        self.sample_by(_sampler)
        return self

    def sample_by(self, sampler: sp.Sampler):
        self._sampler = sampler
        return self

    def batch_sample_by(self, batch_sampler: sp.BatchSampler):
        self._batch_sampler = batch_sampler
        return self

    @property
    def DataLoader(self) -> ClassVar[DataLoader]:
        kwargs = {}
        if self._dataloader_kwargs is not None:
            kwargs.update(self._dataloader_kwargs)

        kwargs['sampler'] = self.sampler
        kwargs['batch_sampler'] = self.batch_sampler

        kwargs['collate_fn'] = CollateBase()

        res = _loader_partial(DataLoader, self, **kwargs)
        return res


class DatasetWrap(BaseBuilder):

    def __init__(self, dataset):
        super().__init__()
        self._dataset = dataset

    def __getitem__(self, index):
        return self._dataset[index]

    def __len__(self):
        return len(self._dataset)

    @classmethod
    def check_then_wrap(cls, dataset: Dataset):
        if isinstance(dataset, BaseBuilder):
            return dataset
        else:
            return cls(dataset)


class PandasBuilder(BaseBuilder):
    def __init__(self, df, white_mode=True):
        super().__init__()
        import pandas as pd
        self.table: pd.DataFrame = df  #
        self.white_mode = white_mode
        self.keys = set()

    @property
    def columns(self):
        return self.table.columns

    def add_white(self, key: str):
        assert key in self.table.columns and self.white_mode
        self.keys.add(key)

    def add_black(self, key: str):
        assert key in self.table.columns and not self.white_mode
        self.keys.add(key)

    def chunk(self, chunksize, drop_last=False):
        assert chunksize > 0
        batch_num = len(self.table) // chunksize
        if drop_last:
            end = chunksize * batch_num
        else:
            end = len(self.table)
        for i in range(0, end, chunksize):
            yield self.table.iloc[i:i + chunksize]

    def __len__(self):
        return len(self.table)

    def __getitem__(self, index):
        res = self.table.iloc[index].to_dict()
        if self.white_mode:
            res = {k: v for k, v in res.item() if k in self.keys}
        else:
            res = {k: v for k, v in res.items() if k not in self.keys}
        return res


class DatasetBuilder(BaseBuilder):
    """
    """

    def __init__(self, xs=None, ys=None, auto_output=False,
                 xs_ipt_transform=None, ys_ipt_transform=None,
                 xs_opt_transform=None, ys_opt_transform=None):
        super().__init__()
        self._dataset_len = None
        self._safe = False
        self._last = None
        self._input_dict = OrderedDict()
        self._output_dict = OrderedDict()
        self._delegate_keys = set()
        self._delegate_outs = OrderedDict()
        self._check_delegate_flag = set()
        self._idnames = []

        self._input_transform = OrderedDict()
        self._output_transform = OrderedDict()
        self._global_transform = []

        if xs is not None:
            self.add_input('xs', xs, xs_ipt_transform)
            if auto_output:
                self.add_output('xs', 'xs', xs_opt_transform)
        if ys is not None:
            self.add_input('ys', ys, ys_opt_transform)
            if auto_output:
                self.add_output('ys', 'ys', ys_opt_transform)

    def __repr__(self):
        return (f"{self.__class__.__name__}("
                f"inputs: {','.join(self._input_dict.keys())}; "
                f"outputs: {','.join(self._output_dict.keys())}; "
                f"size: {len(self)}; "
                f")")

    def _get_item(self, index):
        index = self._map_index(index)

        names = set(self._output_dict.values())

        raw = {}
        for name in names:  # type:str
            source_ = self._input_dict[name]
            sample_ = source_[index]
            raw[name] = sample_

        dele_unpacked_raw = {}
        for dele_name in self._delegate_outs.values():
            source_ = self._input_dict[dele_name]
            if dele_name in self._input_transform:  # transform delegate item directly
                source_ = self._input_transform[dele_name](source_)
            sample_ = source_[index]
            self._check_delegate_res(sample_)
            ress = self._unpack_delegate_res(dele_name, sample_)  # then unpack delegate item
            for res in ress:
                dele_unpacked_raw[res.name] = res.value

        for name, transform in self._input_transform.items():
            if name in raw:
                raw[name] = transform(raw[name])

        sample = OrderedDict()

        [sample.__setitem__(k, index) for k in self._idnames]
        [sample.__setitem__(outkey, raw[name]) for outkey, name in self._output_dict.items()]
        [sample.__setitem__(dele_outkey, dele_item) for dele_outkey, dele_item in dele_unpacked_raw.items()]

        for outkey, transform in self._output_transform.items():
            if outkey in sample:
                sample[outkey] = transform(sample[outkey])

        for transform in self._global_transform:
            sample = transform(sample)

        if self._chain:
            return list(sample.values())
        return sample

    def __getitem__(self, index):
        if self._safe:
            try:
                sample = self._get_item(index)
                self._last = sample
            except BaseException as e:
                import traceback
                from lumo.kit.logger import get_global_logger
                logger = get_global_logger()
                if self._last is not None:
                    except_str = traceback.format_exception(type(e), e, e.__traceback__)
                    logger.error(index, except_str)
                    return self._last
                else:
                    raise e
        else:
            return self._get_item(index)

    @property
    def raw_len(self):
        if self._dataset_len is None:
            return 0
        return self._dataset_len

    def _unpack_delegate_res(self, delegate_name: str,
                             sample_: Union[Data, List[Data], Dict[str, Data]]) -> List[Data]:
        res = []
        if isinstance(sample_, Data):
            if sample_.name is None:
                sample_.name = delegate_name
            res.append(sample_)
        elif isinstance(sample_, list):
            for i, sample__ in enumerate(sample_):
                res.extend(self._unpack_delegate_res(f"{delegate_name}_{i}", sample__))
        elif isinstance(sample_, dict):
            for k, v in sample_.items():
                if isinstance(v, Data):
                    # name will be assigned by dict key when Data.name is None
                    if v.name is not None:
                        v.name = k
                    res.append(v)
                elif isinstance(v, (tuple, list, dict)):  # untreated for deeper level dict
                    res.append(self._unpack_delegate_res(k, v))
                else:  # unwrapped value in dict is allowed, cause we can build it by dict key.
                    res.append(Data(v, name=k))
        return res

    def _check_delegate_res(self, sample_, delegate_name=None):
        """
        To format sample output data as a dict, the basic element of the delegate returned result
        must be wrapped in `Data`. This function will check if every basic elements in sample is a Data instance.
        If not, a `DelegateDataTypeError` Exception will be raised.
        """
        if delegate_name in self._check_delegate_flag:
            return True

        if isinstance(sample_, Data):
            self._check_delegate_flag.add(delegate_name)
            return True
        elif isinstance(sample_, (list, tuple)):
            for i in sample_:
                self._check_delegate_res(i, delegate_name)
            self._check_delegate_flag.add(delegate_name)
            return True
        elif isinstance(sample_, dict):
            for k, v in sample_.items():
                if isinstance(v, (list, dict, tuple)):
                    self._check_delegate_res(v, delegate_name)
            self._check_delegate_flag.add(delegate_name)
            return True
        else:
            raise DelegateDataTypeError(
                f'Delegate must return `Data` object or its list/dict wrap, but got {type(sample_)}.'
            )

    def _check_source_name(self, name, added=True):
        """
        Each source must have different name, or the newer will replace the older.
        """
        if added:
            assert name not in self._input_dict, f'source "{name}" has been added'
        else:
            assert name in self._input_dict, f'source "{name}" has not been added'

    def _check_outkey_name(self, outkey, added=True):
        """
        Each output key must be different, or new output will cover the old one.
        """
        if added:
            assert outkey not in self._output_dict, f'outkey "{outkey}" has been added'
        else:
            assert outkey in self._output_dict, f'outkey "{outkey}" has not been added'

    def _map_index(self, index):
        """
        Map the raw index to the final index for source data.
        Args:
            index:

        Returns:

        """
        if self._sub_indices is not None:
            index = self._sub_indices[index]
        if self._reindices is not None:
            index = self._reindices[index]

        return index

    def add_ids(self, name):
        assert name not in self._idnames, f'id key "{name}" has been added'
        self._idnames.append(name)
        return self

    def add_input(self, name, source: Union[Sized, DataDelegate], transform=None):
        """
        Add an source/input stream.
        Args:
            name:
            source:
        """
        self._check_len(source)
        self._check_source_name(name, added=True)
        self._input_dict[name] = source
        if isinstance(source, DataDelegate):
            self._delegate_keys.add(name)

        if transform is not None:
            self.add_input_transform(name, transform)
        return self

    def add_output(self, name, outkey=None, transform=None, in_delegate=False):
        """
        Add an output stream.
        Args:
            name:
            outkey:
        """

        if not in_delegate:
            assert outkey is not None
            self._check_source_name(name, added=False)
            self._check_outkey_name(outkey)
            self._output_dict[outkey] = name
            if transform is not None:
                self.add_output_transform(outkey, transform)
        else:
            if outkey is None:
                outkey = f"dele_{len(self._delegate_outs)}"
            self._delegate_outs[outkey] = name
        return self

    def add_input_transform(self, name, transform: Callable[[Any], Any]):
        """
        If there are multiple outputs want to fetch data from a same source, and there is a function that
        only need to be called once, then you can add an `input transform` to achieve it.

        Args:
            name:
            transform:

        Notes:
              One input source with name `name` will only have one input transform, an Exception will be raised
              when added second one with the same `name`. So combine your transforms when you have multiple transforms
              for the `name` source.
        """
        self._check_source_name(name, added=False)
        self._input_transform[name] = transform
        return self

    def add_output_transform(self, outkey, transform: Callable[[Any], Any]):
        """
        When you add an `output transform` with `outkey`, it means the output data with name `outkey`
        will be passed into this transform by call `transform(data)`

        Args:
            outkey:
            transform: callable object

        Notes:
              One output data with name `outkey` will only have one transform, an Exception will be raised
              when added second one with the same `outkey`. So combine your transforms when you have multiple transforms
              for the `outkey` data.
        """
        self._check_outkey_name(outkey, added=False)
        self._output_transform[outkey] = transform
        return self

    def add_global_transform(self, transform: Callable[[Dict[str, Any]], Any]):
        self._global_transform.append(transform)
        return self

    def safe(self):
        self._safe = True
        return self

    def copy(self):
        builder = super().copy()

        builder._dataset_len = self._dataset_len
        builder._input_dict = OrderedDict(self._input_dict)
        builder._output_dict = OrderedDict(self._output_dict)
        builder._idnames = list(self._idnames)

        builder._input_transform = OrderedDict(self._input_transform)
        builder._output_transform = OrderedDict(self._output_transform)
        builder._global_transform = list(self._global_transform)

        return builder

    @property
    def sampler(self):
        return self._sampler

    @property
    def batch_sampler(self):
        return self._batch_sampler


class RemoteDataset(BaseBuilder):
    def as_client(self, ds_name, address='127.0.0.1', port=9638):
        pass

    def as_server(self, ds_name, address='127.0.0.1', port=9638):
        import fastapi
        api = fastapi.FastAPI()

        return api


class SimpleDataset(DatasetBuilder):
    def __init__(self,
                 xs=None, ys=None,
                 transform=None, target_transform=None, global_transform=None,
                 indices: Sequence[int] = None, subset: Sequence[int] = None,
                 chain=False):
        super().__init__()
        if xs is not None:
            self.add_input(name='xs', source=xs).add_output(name='xs', outkey='xs')
            if transform is not None:
                self.add_output_transform('xs', transform)

        if ys is not None:
            self.add_input(name='ys', source=ys).add_output(name='ys', outkey='ys')
            if target_transform is not None:
                self.add_output_transform('ys', target_transform)

        if global_transform is not None:
            self.add_global_transform(global_transform)

        if indices is not None:
            self.reindices(indices)

        if subset is not None:
            self.subset(subset)

        if chain:
            self.chain()


class ConcatDataset(BaseBuilder):
    def __init__(self, *builders: DatasetBuilder):
        self._check_builder_iter_type(*builders)
        self._chunk = builders
        super().__init__()

    def _check_builder_iter_type(self, *args: DatasetBuilder):
        assert sum([i._chain for i in args]) in {0, len(args)}
        self._chain = args[0]._chain

    def __repr__(self):
        return (f"{self.__class__.__name__}("
                f"Chains: {','.join([str(i) for i in self._chunk])}; "
                f"Size: {len(self)}; "
                f")")


class ChainDataset(ConcatDataset):

    def __getitem__(self, index):
        index = self._map_index(index)
        lens = [len(i) for i in self._chunk]

        for i, offset in enumerate(accumulate(lens, add)):
            if index < offset:
                return self._chunk[index - lens[i]]

    @property
    def raw_len(self):
        return sum(len(i) for i in self._chunk)


class ZipDataset(ConcatDataset):
    def __init__(self, *builders: BaseBuilder):
        super().__init__(*builders)
        self._counter = [0 for _ in builders]

    def __len__(self):
        return super().__len__()

    @property
    def raw_len(self):
        return max(len(i) for i in self._chunk)

    def __getitem__(self, index):
        index = self._map_index(index)
        lens = [len(i) for i in self._chunk]

        n_counter = []
        ress = []
        for count, builder, len_ in zip(self._counter, self._chunk, lens):
            if index >= len_:
                index = ((index + count) % len_)
                count = (count + 1) % len_
            res = builder[index]
            ress.append(res)
            n_counter.append(count)

        self._counter = n_counter

        if not self._chain:
            nress = {}
            for res in ress:
                for k, v in res.items():
                    nress[k] = v
            ress = nress
        return ress



import fire

doc = """
Usage:
# create templete directory 
lumo init [dir]

# easier way to open tensorboard 
# lumo board [--logdir=<logdir>]
# lumo board [--test=<test_name>] # find test_name and tensorboard it 
# lumo board  # default open ./board

# restore code snapshot of some test
lumo reset <test_name>

# archive code snapshot of some test
lumo archive <test_name>

# print log file
lumo log <test_name>

# print params of this test
lumo params <test_name>

# <test_name>/--test=<test_name>/--test_name=<test_name>

# TODO
lumo config local --k=v
lumo config global --k=v

# get a free port
lumo port

"""
import sys
from lumo import __version__

from lumo.decorators import regist_func_to
from lumo import globs

func_map = {}


# @regist_func_to(func_map)
def init(src_dir=None, template=None):
    import shutil
    import os

    if src_dir is None:
        src_dir = os.getcwd()
    from .api.git import init_repo

    # dir_name = os.path.basename(src_dir)
    _, init = init_repo(src_dir)
    if init:
        print(f'{src_dir} initialized')
    else:
        print(f'{src_dir} is already a git dir.')

    # templete_dir = os.path.join(os.path.dirname(__file__), 'templete')
    # shutil.copytree(templete_dir, os.path.join(src_dir, dir_name))
    # # os.rename(os.path.join(src_dir, 'templete'), os.path.join(src_dir, dir_name))
    # for root, dirs, files in os.walk(src_dir):
    #     for file in files:
    #         if file.endswith('.py-tpl'):
    #             nfile = "{}.py".format(os.path.splitext(file)[0])
    #             os.rename(os.path.join(root, file), os.path.join(root, nfile))


#
def _board_with_logdir(logdir, *args, **kwargs):
    import os
    import subprocess

    tmpdir = os.path.join(os.path.dirname(logdir), 'board_tmp')
    os.makedirs(tmpdir, exist_ok=True)
    try:
        subprocess.check_call(['tensorboard', '--logdir={}'.format(logdir),
                               *['--{}={}'.format(k, v) for k, v in kwargs.items()]],
                              env=dict(os.environ, TMPDIR=tmpdir))
    except KeyboardInterrupt as k:
        print('bye.')
        exit(0)


def _board_with_test_name(test_name, *args, **kwargs):
    query = Q.tests(test_name)
    if query.empty:
        raise IndexError(test_name)
    vw = query.to_viewer()
    if not vw.has_board():
        raise AttributeError('{} has no board or has been deleted'.format(test_name))
    else:
        _board_with_logdir(vw.board_logdir, **kwargs)


@regist_func_to(func_map)
def board(*args, **kwargs):
    if len(args) > 0:
        kwargs.setdefault('test_name', args[0])

    if 'logdir' in kwargs:
        _board_with_logdir(**kwargs)
    elif 'test' in kwargs:
        _board_with_test_name(kwargs['test'])
    elif 'test_name' in kwargs:
        _board_with_test_name(**kwargs)
    else:
        _board_with_logdir('./board')


def _find_test_name(*args, **kwargs):
    if len(args) > 0:
        return args[0]
    elif 'test' in kwargs:
        return kwargs['test']
    elif 'test_name' in kwargs:
        return kwargs['test_name']
    return None


@regist_func_to(func_map)
def reset(*args, **kwargs):
    test_name = _find_test_name(*args, **kwargs)
    query = Q.tests(test_name)
    if query.empty:
        print("can't find test {}".format(test_name))
        exit(1)
    exp = query.to_viewer().reset()
    print('reset from {} to {}'.format(exp.plugins['reset']['from'], exp.plugins['reset']['to']))


#
@regist_func_to(func_map)
def archive(*args, **kwargs):
    test_name = _find_test_name(*args, **kwargs)
    query = Q.tests(test_name)
    if query.empty:
        print("can't find test {}".format(test_name))
        exit(1)
    exp = query.to_viewer().archive()

    print('archive {} to {}'.format(test_name, exp.plugins['archive']['file']))


@regist_func_to(func_map)
def log(*args, **kwargs):
    test_name = _find_test_name(*args, **kwargs)
    query = Q.tests(test_name)
    if query.empty:
        print("can't find test {}".format(test_name))
        exit(1)

    vw = query.to_viewer()
    if not vw.has_log():
        print("can't find log file of [{}]".format(test_name))
        exit(1)
    print(vw.log_fn)


#
#
@regist_func_to(func_map)
def params(*args, **kwargs):
    test_name = _find_test_name(*args, **kwargs)
    query = Q.tests(test_name)
    if query.empty:
        print("can't find test {}".format(test_name))
        exit(1)

    vw = query.to_viewer()
    p = vw.params
    if p is not None:
        print(p)
    else:
        print("can't find param object of [{}]".format(test_name))
        exit(1)


def config(field=None, k=None, v=None):
    """

    Args:
        k: key of the value
        v: value for key
    """
    if field is None or k is None:
        print(config.__doc__)

    if field == 'list':
        print(globs)
        return

    if v is None:
        v = globs.get(k, level=field, default=None)
        if v is None:
            print(f"{k} not exist in field {field}.")
        else:
            print(f'{k} = {v}')
        return
    globs.set(k, v, level=field)


def main(*args, **kwargs):
    # print(args, kwargs)
    print(f"lumo {__version__}")
    if len(args) == 0 or 'help' in kwargs:
        print(doc)
        return

    branch = args[0]
    if branch in func_map:
        func_map[branch](*args[1:], **kwargs)
    else:
        print(doc)


# Fire 不能嵌套在 __main__ 判断里，否则 sys.argv 识别会出问题，目前原因未知
fire.Fire(main)
exit(0)



"""

"""


import sys
import time
from lumo import Logger
logger = Logger()
for i in range(1000):
    time.sleep(0.5)
    # sys.stdin.buffer

    # buf = ''
    print('value')
    buf = sys.stdin.read()
    if len(buf) > 0:
        print(buf)



"""
Methods about file/date/format.
"""
import os
from datetime import datetime
import hashlib


def strftime(fmt='%y-%m-%d-%H%M%S', dateobj: datetime = None):
    """get current date with formatted"""
    if dateobj is not None:
        return dateobj.strftime(fmt)
    return datetime.now().strftime(fmt)


def date_from_str(value, fmt='%y-%m-%d-%H%M%S') -> datetime:
    """convert from date formatted string to datetime object"""
    return datetime.strptime(value, fmt)


def file_atime2date(file, fmt='%y%m%d-%H%M%S') -> str:
    """
    get atime of a file and convert in from timestamp to formatted string
    atime : time of last access
    """
    return strftime(fmt, datetime.fromtimestamp(os.path.getatime(file)))


def file_mtime2date(file, fmt='%y%m%d-%H%M%S') -> str:
    """
    get mtime of a file and convert in from timestamp to formatted string
    mtime : time of last modification (write)
    """
    return strftime(fmt, datetime.fromtimestamp(os.path.getmtime(file)))


def file_ctime2date(file, fmt='%y%m%d-%H%M%S') -> str:
    """
    get ctime of a file and convert in from timestamp to formatted string
    ctime : time of last status change
    """
    return strftime(fmt, datetime.fromtimestamp(os.path.getctime(file)))


def timehash():
    import time
    time_ns = str(time.time_ns())
    hl = hashlib.md5()
    hl.update(time_ns.encode(encoding='utf-8'))
    return hl.hexdigest()



"""

"""
import time

import psutil

from lumo.kit.experiment import Experiment
from lumo.proc.const import EXP_CONST
from lumo.kit.params import BaseParams
from lumo.proc.date import strftime


def wait_pid_stop():
    params = BaseParams()
    params.pid = None
    params.test_name = None
    params.exp_name = None
    params.from_args()

    exp = Experiment(params.exp_name, params.test_name)
    c = 0
    while psutil.pid_exists(params.pid):
        info = exp.load_info(EXP_CONST.INFO_KEY.STATE)
        if 'end_code' in info:
            break

        exp.dump_info(EXP_CONST.INFO_KEY.STATE, {
            'end': strftime(),
        }, append=True)
        time.sleep(1)


if __name__ == '__main__':
    wait_pid_stop()



"""

"""


class tree(dict):
    """Implementation of perl's autovivification feature."""

    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            value = self[item] = type(self)()
            return value

    def walk(self):
        for k, v in self.items():
            yield k, v
            if isinstance(v, tree):
                for kk, vv in v.walk():
                    yield f'{k}/{kk}', vv



from ..kit.params import BaseParams
from ..base_classes.params_vars import OptimBuilder


class DataLoaderPM(BaseParams):
    def __init__(self):
        super().__init__()
        self.num_workers = 4
        self.batch_size = 12
        self.pin_memory = True


class EvalFirstPM(BaseParams):
    def __init__(self):
        super().__init__()
        self.eval_first = False


class OptimPM(BaseParams):
    def __init__(self):
        super().__init__()
        self.optim = None  # type:OptimBuilder



from lumo.calculate.schedule import *

if __name__ == '__main__':
    lsc = LinearSchedule(left=0,right=10,start=1,end=0)
    sched = PeriodHalfCosSchedule(left=0, right=100, start=1, end=0)
    from matplotlib import pyplot as plt
    # lsc.plot()
    sched.plot(left=-100,right=300)
    plt.show()
    print(sched)


from lumo import Saver
import time
import shutil


def test_save_load():
    save_root = './temp_saver'
    saver = Saver(save_root)
    epoch = 20
    max_keep = 5
    for i in range(epoch):
        time.sleep(0.01)
        saver.save_model(i, {'step': i}, {'meta_step': i}, is_best=(i == 5))
        if i % 2 == 0:
            saver.save_keypoint(i, {'step': i}, {'meta_step': i})
        else:
            saver.save_checkpoint(i, {'step': i}, {'meta_step': i}, max_keep=max_keep, is_best=(i % 5) == 0)

    assert len(saver.list_models()) == (epoch)

    state = saver.load_model(best_if_exist=True, with_meta=True)
    assert state[0]['step'] == 5 and state[1]['meta_step'] == 5

    state = saver.load_model(2)
    assert state['step'] == 2

    assert len(saver.list_keypoints()) == (epoch // 2)
    state = saver.load_checkpoint(best_if_exist=True)
    assert state['step'] == 15

    state = saver.load_checkpoint(0)  # max keep=5 ,11 , 13, 15, 17, 19 for it.
    assert state['step'] == 11

    assert len(saver.list_checkpoints()) == max_keep
    state = saver.load_keypoint(2)
    assert state['step'] == 4

    shutil.rmtree(save_root)



import os
import stat
import sys
from lumo.utils.exithook import wrap_before, wrap_after
from lumo.proc.path import libhome, local_dir
from .experiment import Experiment
from ..proc.const import EXP_CONST, FN


class ExpHook():
    def regist(self, exp: Experiment): self.exp = exp

    def on_start(self, exp: Experiment): pass

    def on_end(self, exp: Experiment): pass


class LastCmd(ExpHook):
    def on_start(self, exp: Experiment):
        with open('lastcmd.sh', 'w', encoding='utf-8') as w:
            w.write(' '.join(exp.exec_argv))

        st = os.stat('lastcmd.sh')
        os.chmod('lastcmd.sh', st.st_mode | stat.S_IEXEC)


class LogCmd(ExpHook):
    """a './cache/cmds.log' file will be generated, """

    def on_start(self, exp: Experiment):
        from lumo.proc.date import strftime
        fn = exp.project_cache_fn(f'{strftime("%y-%m-%d")}.log', 'cmds')
        res = exp.exec_argv

        with open(fn, 'a', encoding='utf-8') as w:
            w.write(f'{strftime("%H:%M:%S")}, {exp.test_root}, {res[0]}, {exp.commit_hash}\n')
            res[0] = os.path.basename(res[0])
            w.write(f"> {' '.join(res)}")
            w.write('\n\n')


class LogTestGlobally(ExpHook):
    def on_start(self, exp: Experiment):
        fn = os.path.join(libhome(), FN.TESTLOG)
        with open(fn, 'a', encoding='utf-8') as w:
            w.write(f'{exp.test_root}\n')


class LogTestLocally(ExpHook):
    def on_start(self, exp: Experiment):
        local_ = local_dir()
        if local_ is None:
            return
        fn = os.path.join(local_, FN.TESTLOG)
        with open(fn, 'a', encoding='utf-8') as w:
            w.write(f'{exp.test_root}\n')


class RegistRepo(ExpHook):
    def on_start(self, exp: Experiment):
        from ..proc.const import FN
        from ..proc.const import CFG
        from lumo.utils import safe_io as io
        fn = os.path.join(libhome(), FN.REPOSJS)
        res = None
        if os.path.exists(fn):
            res = io.load_json(fn)
        if res is None:
            res = {}

        inner = res.setdefault(exp.project_hash, {})
        inner['name'] = exp.project_name
        repos = inner.setdefault('repo', [])
        if exp.project_root not in repos:
            repos.append(exp.project_root)
        storages = inner.setdefault('exp_root', [])
        if exp.exp_root not in storages:
            storages.append(exp.exp_root)

        io.dump_json(res, fn)


class RecordAbort(ExpHook):
    def __init__(self):
        wrap_before(self.exc_end)

    def exc_end(self, exc_type, exc_val, exc_tb):
        import traceback
        res = traceback.format_exception(exc_type, exc_val, exc_tb)
        res = [i for i in res if 'in _newfunc' not in i]
        self.exp.writeline('exception', "".join(res))
        self.exp.end(
            end_code=1,
            exc_type=traceback.format_exception_only(exc_type, exc_val)[-1].strip()
        )


class PrintExpId(ExpHook):

    def regist(self, exp: Experiment):
        super().regist(exp)
        self.exp.add_exit_hook(self.on_exit)

    def on_exit(self, *args, **kwargs):
        print(f"END TEST {self.exp.short_uuid}")


class LogCMDAndTest(ExpHook):
    def on_start(self, exp: Experiment):
        from lumo.kit.logger import get_global_logger
        # get_global_logger().raw(f"{exp.test_root} | {' '.join(sys.argv)}")

    def on_end(self, exp: Experiment):
        from lumo.kit.logger import get_global_logger
        get_global_logger().raw(f"{exp.test_root} | {' '.join(sys.argv)}")



"""

"""
__version__ = "0.1.9"

from .utils.keys import K

from .kit import (
    F, Q,
    BaseParams, Params,
    Meter, AvgMeter,

    DatasetBuilder, DataDelegate, DataModule,
    Saver, Random,

    Logger,

    TrainerExperiment, exphook,
    globs, Trainer, callbacks
)

from .contrib.data import collate

from . import calculate  # initialize schedule attr classes




from torch import nn
import torch
from copy import deepcopy
from torch import LongTensor

if torch.cuda.is_available():
    from torch.cuda import LongTensor as CLongTensor
else:
    CLongTensor = LongTensor

from typing import TypeVar

Module = TypeVar('Module', bound=nn.Module)


def EMA(model: Module, alpha=0.999) -> Module:
    """
    Exponential Moving Average(EMA) for nn.Module
    Args:
        model: nn.Module, An EMA wrapper of original model
        alpha: float, default 0.999, decay ratio of EMA

    Returns:
        A new cloned model that has a new method 'step'

    Notes:
        ema model will not generate gradient in its forward process
    """
    ema_model = deepcopy(model)
    [i.detach_() for i in ema_model.parameters()]

    def step(alpha_=None):

        if alpha_ is None:
            alpha_ = alpha

        with torch.no_grad():
            for (_, ema_param), (_, param) in zip(ema_model.state_dict().items(), model.state_dict().items()):
                ema_param.to(param.device)
                if not isinstance(param, (LongTensor, CLongTensor)):
                    ema_param.data.mul_(alpha_).add_(param.data, alpha=1 - alpha_)
                else:
                    ema_param.data.copy_(alpha_ * ema_param + (1 - alpha_) * param)

    forward_ = ema_model.forward

    def forward(*args, **kwargs):
        with torch.no_grad():
            return forward_(*args, **kwargs)

    ema_model.forward = forward
    ema_model.step = step
    return ema_model



from functools import lru_cache
from .const import ENV


@lru_cache(1)
def git_enable():
    if not ENV.IS_GIT_ENABLED:
        return False

    try:
        import git
    except ImportError:
        return False
    try:
        git.Git().execute(['git', 'rev-parse', '--git-dir'])
        return True
    except git.GitCommandError:
        return False


@lru_cache(1)
def matplotlib_enable():
    try:
        import matplotlib
        return True
    except ImportError:
        return False


@lru_cache(1)
def numpy_enable():
    try:
        import numpy
        return True
    except ImportError:
        return False


@lru_cache(1)
def pandas_enable():
    try:
        import pandas
        return True
    except ImportError:
        return False


@lru_cache(1)
def accelerate_enable():
    try:
        import accelerate
        return True
    except ImportError:
        return False



"""

"""

# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Keras integration for TensorBoard hparams.

Most users should use `tensorboard.plugins.hparams.api` to access this
module's contents.
"""

# from __future__ import absolute_import
# from __future__ import division
# from __future__ import print_function

import six
import tensorflow as tf

from tensorboard.plugins.hparams import api_pb2
from tensorboard.plugins.hparams import summary
from tensorboard.plugins.hparams import summary_v2


class Callback(tf.keras.callbacks.Callback):
  """Callback for logging hyperparameters to TensorBoard.

  NOTE: This callback only works in TensorFlow eager mode.
  """

  def __init__(self, writer, hparams):
    """Create a callback for logging hyperparameters to TensorBoard.

    As with the standard `tf.keras.callbacks.TensorBoard` class, each
    callback object is valid for only one call to `model.fit`.

    Args:
      writer: The `SummaryWriter` object to which hparams should be
        written, or a logdir (as a `str`) to be passed to
        `tf.summary.create_file_writer` to create such a writer.
      hparams: A `dict` mapping hyperparameters to the values used in
        this session. Keys should be the names of `HParam` objects used
        in an experiment, or the `HParam` objects themselves. Values
        should be Python `bool`, `int`, `float`, or `string` values,
        depending on the type of the hyperparameter.

    Raises:
      ValueError: If two entries in `hparams` share the same
        hyperparameter name.
    """
    # Defer creating the actual summary until we write it, so that the
    # timestamp is correct. But create a "dry-run" first to fail fast in
    # case the `hparams` are invalid.
    self._hparams = dict(hparams)
    summary_v2.hparams_pb(self._hparams)
    if writer is None:
      raise TypeError("writer must be a `SummaryWriter` or `str`, not None")
    elif isinstance(writer, str):
      self._writer = tf.compat.v2.summary.create_file_writer(writer)
    else:
      self._writer = writer

  def _get_writer(self):
    if self._writer is None:
      raise RuntimeError(
          "hparams Keras callback cannot be reused across training sessions"
      )
    if not tf.executing_eagerly():
      raise RuntimeError(
          "hparams Keras callback only supported in TensorFlow eager mode"
      )
    return self._writer

  def on_train_begin(self, logs=None):
    del logs  # unused
    with self._get_writer().as_default():
      summary_v2.hparams(self._hparams)

  def on_train_end(self, logs=None):
    del logs  # unused
    with self._get_writer().as_default():
      pb = summary.session_end_pb(api_pb2.STATUS_SUCCESS)
      raw_pb = pb.SerializeToString()
      tf.compat.v2.summary.experimental.write_raw_pb(raw_pb, step=0)
    self._writer = None





"""
Experiment
 - 记录各个参数
 - 生成各个路径
 - 实时记录
"""
import os
import time
import random
import pickle
import sys
import traceback
from typing import Union, TYPE_CHECKING
from uuid import uuid1
from lumo.base_classes import attr
from lumo.kit.environ import globs

from lumo.utils.paths import checkpath
from lumo.utils import safe_io as io
from lumo.proc.date import strftime
from lumo.proc.dist import local_rank, is_dist
from lumo.proc.const import CFG, FN, EXP_CONST
from lumo.proc.path import local_dir
from lumo.proc.explore import git_enable

# from lumo.utils.repository import commit as git_commit
# from lumo.utils.repository import _commits_map

if TYPE_CHECKING:
    from .exphook import ExpHook


class Experiment:
    """
    每一个经由 experiment 辅助的实验，都会在以下几个位置存有记录，方便回溯：
     - ~/.lumo/
     - <working_dir>/.lumo/
     - <save_dir>/ # 记录相对路径

    TODO
    主要功能：
     - 以实验为单位，维护在该实验上做的全部试验，用于提供该次"试验唯一"的目录，该"实验唯一"的目录，"用户唯一"的目录
     - 以试验为单位，存储试验过程中提交到 Experiment 的全部信息（配置信息等）

    主要功能：
     - 以试验为单位，在试验开始前提交本次运行快照

    其中，存储目录从用户配置（本地）读取，但也可以优先通过环境变量控制

    根目录 -> 实验目录 -> 试验目录 -> 试验目录下子目录

    使用方法：

    with Experiment(exp_name) as exp:

    """

    def __init__(self, exp_name, test_name=None):
        self._hooks = []
        self._exp_name = exp_name
        self._test_name = test_name

        from .exphook import RegistRepo, RecordAbort, LogTestLocally, LogTestGlobally
        (
            self.add_hook(RegistRepo())
                .add_hook(RecordAbort())
                .add_hook(LogTestLocally())
                .add_hook(LogTestGlobally())
        )

    def _create_dir(self, root, dirname):
        res = checkpath(root, dirname)
        return res

    def _create_fn(self, basename, dirname, root):
        if dirname is not None:
            res = checkpath(root, dirname)
        else:
            res = root
        return os.path.join(res, basename)

    def _create_agent(self):
        import subprocess, sys
        from lumo.kit import agent
        cmd = [
            sys.executable, '-m', agent.__spec__.name,
            f"--pid={os.getpid()}",
            f"--test_name={self.test_name}",
            f"--exp_name={self.exp_name}",
            # f"--params={sys.argv}" # TODO add sys.argv
        ]
        subprocess.Popen(' '.join(cmd),
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True,
                         start_new_session=True)

    def cache_dir(self, dirname):
        return self._create_dir(self.cache_root, dirname)

    def cache_fn(self, name, dirname=None):
        return self._create_fn(name, dirname, self.cache_root)

    def project_cache_fn(self, name, dirname=None):
        return self._create_fn(name, dirname, self.project_cache_root)

    def project_cache_dir(self, dirname):
        return self._create_dir(self.project_cache_root, dirname)

    def exp_dir(self, name):
        return self._create_dir(self.exp_root, name)

    def project_dir(self, name):
        return self._create_dir(self.project_root, name)

    def project_fn(self, basename, dirname=None):
        return self._create_fn(basename, dirname, self.project_root)

    def exp_fn(self, basename, dirname=None):
        return self._create_fn(basename, dirname, self.exp_root)

    def test_dir(self, name):
        return self._create_dir(self.test_root, name)

    def test_fn(self, basename, dirname=None):
        return self._create_fn(basename, dirname, self.test_root)

    def load_info(self, key):
        fn = self.test_fn(self._create_info_basename(key), EXP_CONST.IO_DIR.SINFO_DIR)
        if not os.path.exists(fn):
            return None
        return attr.from_dict(io.load_json(fn))

    def dump_info(self, key, info: dict, append=False):
        info = attr(info).jsonify()
        fn = self.test_fn(self._create_info_basename(key), EXP_CONST.IO_DIR.SINFO_DIR)
        if append and os.path.exists(fn):
            old_info = self.load_info(key)
            for k, v in info.items():
                old_info[k] = v
            info = old_info
        io.dump_json(info, fn)
        return fn

    def load_pkl(self, key):
        fn = self.test_fn(self._create_bin_basename(key), EXP_CONST.IO_DIR.PKL_DIR)
        if not os.path.exists(fn):
            return None
        with open(fn, 'r') as r:
            return pickle.load(r)

    def dump_pkl(self, key, info):
        fn = self.test_fn(self._create_bin_basename(key), EXP_CONST.IO_DIR.PKL_DIR)
        with open(fn, 'wb') as w:
            pickle.dump(info, w)
        return fn

    def writeline(self, key, value: str):
        fn = self.test_fn(f"{key}.txt", EXP_CONST.IO_DIR.INFO_DIR)
        with open(fn, 'w', encoding='utf-8') as w:
            w.write(value)

    def readline(self, key):
        fn = self.test_fn(f"{key}.txt", EXP_CONST.IO_DIR.INFO_DIR)
        if not os.path.exists(fn):
            return ''
        with open(fn, 'r', encoding='utf-8') as r:
            return ''.join(r.readlines())

    def readlines(self, raw=False) -> Union[dict, str]:
        line_root = os.path.join(self.test_root, EXP_CONST.IO_DIR.INFO_DIR)
        fs_ = os.listdir(line_root)
        fs_ = [f for f in fs_ if f.endswith('.txt')]
        res = {}
        for f in fs_:
            key = f[:-1]
            line = self.readline(key)
            res[key] = line
        if raw:
            return res
        else:
            strs = []
            for k, v in res.items():
                strs.append(k)
                strs.append('-' * len(k))
                strs.append(v)
                strs.append('=' * len(k))
            return '\n'.join(strs)

    def _create_bin_basename(self, key):
        return f"{key}.pkl"

    def _create_info_basename(self, key):
        return f"{key}.json"

    def __enter__(self):
        return self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        extra = {}
        if exc_type is not None:
            exc_type = traceback.format_exception_only(exc_type, exc_val)[-1].strip()
            extra['exc_type'] = exc_type
            extra['end_info'] = str(exc_type)
            extra['end_code'] = 1
            self.writeline('exception',
                           "".join(traceback.format_exception(exc_type, exc_val, exc_tb)))
        self.end(**extra)

    @property
    def exec_argv(self):
        return [sys.executable, *sys.argv]

    @property
    def project_hash(self):
        hash_fn = os.path.join(self.project_root, FN.PHASH)
        if os.path.exists(hash_fn):
            return io.load_string(hash_fn)
        else:
            with open(hash_fn, 'w') as w:
                hash = uuid1().hex[-8:-4]
                w.write(hash)
            return hash

    @property
    def project_name(self):
        return os.path.basename(self.project_root)

    @property
    def exp_name(self):
        return self._exp_name

    @property
    def test_name(self):
        """Create unique name for the current test"""

        def _create_test_name():
            from lumo.proc.date import strftime, timehash
            fs = os.listdir(self.exp_root)
            date_str = strftime('%y%m%d')
            fs = [i for i in fs if i.startswith(date_str)]
            _test_name = f"{date_str}.{len(fs):03d}.{timehash()[-6:-4]}t"
            return _test_name

        if self._test_name is None:
            if is_dist():
                flag_fn = f'.{os.getppid()}'
                if local_rank() > 0:
                    time.sleep(random.randint(2, 4))
                    fn = self.exp_fn(flag_fn)
                    if os.path.exists(fn):
                        with open(fn, 'r') as r:
                            self._test_name = r.readline().strip()
                else:
                    self._test_name = _create_test_name()
                    fn = self.exp_fn(flag_fn)
                    with open(fn, 'w') as w:
                        w.write(self._test_name)
            else:
                self._test_name = _create_test_name()

        return self._test_name

    @property
    def storage_root(self):
        """experiments root dir, default is `~/.lumo/experiments`"""
        path = globs[CFG.PATH.GLOBAL_EXP]

        return checkpath(path)

    @property
    def cache_root(self) -> str:
        """cache root for lumo, default"""
        res = globs[CFG.PATH.CACHE]
        return checkpath(res)

    @property
    def exp_root(self) -> str:
        """root dir for current experiment"""
        return checkpath(self.storage_root, self.exp_name)

    @property
    def test_root(self) -> str:
        """Root dir for current test"""
        return checkpath(self.exp_root, self.test_name)

    @property
    def project_root(self) -> str:
        """git repository root(working dir)"""
        return local_dir()

    @property
    def project_cache_root(self) -> str:
        """<project_root>/.cache or .cache/lumo"""
        return checkpath(self.project_root, '.cache')

    @property
    def uuid(self):
        return self.readline('uuid')

    @property
    def short_uuid(self):
        return self.uuid[:6]

    @property
    def commit_hash(self) -> str:
        from lumo.utils.repository import _commits_map
        res = _commits_map.get('lumo', None)
        if res is not None:
            return res.hexsha[:8]
        return ''

    def dump_experiment_info(self):
        self.dump_info(EXP_CONST.INFO_KEY.STATE, {
            'start': strftime(),
            'end': strftime()
        })

        self.dump_info(EXP_CONST.INFO_KEY.EXECUTE, {
            'repo': self.project_root,
            'cwd': os.getcwd(),
            'exec_file': sys.argv[0],
            'exec_bin': sys.executable,
            'exec_argv': sys.argv
        })

        self.dump_info(EXP_CONST.INFO_KEY.PROJECT, {
            'hash': self.project_hash,
            'root': self.project_root,
        })

        if git_enable() and local_rank() <= 0:
            from lumo.utils.repository import commit
            commit_ = commit(key='lumo', info=self.test_root)
            commit_hex = commit_.hexsha[:8]
            if commit_ is not None:
                self.writeline('commit', commit_hex)
                self.dump_info(EXP_CONST.INFO_KEY.GIT, {
                    'commit': commit_hex,
                    'repo': self.project_root,
                })

        self.writeline('uuid', uuid1().hex)

        from lumo import __version__
        self.dump_info(EXP_CONST.INFO_KEY.VERSION, {
            'lumo': __version__,
        })

    def start(self):
        self._create_agent()
        self.dump_experiment_info()
        for hook in self._hooks:  # type: ExpHook
            hook.on_start(self)
        return self

    def end(self, enc_code=0, **extra):
        self.dump_info(EXP_CONST.INFO_KEY.STATE, {
            'end_code': enc_code,
            **extra,
        }, append=True)

        for hook in self._hooks:  # type: ExpHook
            hook.on_end(self)
        return self

    def add_hook(self, hook: 'ExpHook'):
        hook.regist(self)
        self._hooks.append(hook)
        return self

    def add_exit_hook(self, func):
        import atexit
        def exp_func():
            func(self)

        atexit.register(exp_func)


class TrainerExperiment(Experiment):

    def __init__(self, exp_name, test_name=None):
        super().__init__(exp_name, test_name)
        from .exphook import LastCmd, LogCmd, LogCMDAndTest
        self.add_hook(LastCmd())
        self.add_hook(LogCmd())
        self.add_hook(LogCMDAndTest())

    class VAR_KEY:
        WRITER = 'board'

    @property
    def log_dir(self):
        return self.test_root

    @property
    def params_fn(self):
        res = self.test_fn('params.json')
        self.writeline('params', res)
        return res

    @property
    def board_args(self):
        log_dir = self.test_dir('board')
        self.writeline('writer', log_dir)
        return {
            'filename_suffix': '.bd',
            'log_dir': log_dir,
        }

    @property
    def saver_dir(self):
        res = self.test_dir('saver')
        self.writeline('saver', res)
        return res

    @property
    def rnd_dir(self):
        res = self.project_cache_dir('rnd')
        self.writeline('rnd', res)
        return res

    def dump_experiment_info(self):
        super().dump_experiment_info()
        try:
            import torch
            if torch.cuda.is_available():
                cuda_version = torch.cuda_version
            else:
                cuda_version = '0'
            self.dump_info(EXP_CONST.INFO_KEY.VERSION, {
                'torch': torch.__version__,
                'cuda': cuda_version
            }, append=True)
        except:
            pass

    def dump_train_info(self, epoch):
        self.dump_info(EXP_CONST.INFO_KEY.TRAINER, {
            'epoch': epoch
        }, append=True)



"""

"""
from typing import List, Union, Tuple

import torch



def classify(preds: torch.Tensor,
             labels: torch.Tensor,
             cacu_rate: bool = False,
             topk: List[int] = None) -> Union[List[float], Tuple[int, List[int]]]:
    """
    用于在分类问题中计算准确率

    Args:
        preds: [batch,logits]
        labels: [labels,]
        cacu_rate: 计算正确率而不是计数
        topk: list(int) ，表明计算哪些topk，默认计算 top1 和 top5

    Returns:
        if cacu_rate:
            [topk_rate,...]
        else:
            total, [topk_count,...]
    """
    if topk is None:
        topk = (1, 5)
    k = topk
    _, maxk = torch.topk(preds, max(*k), dim=-1)
    total = labels.size(0)
    test_labels = labels.view(-1, 1)  # reshape labels from [n] to [n,1] to compare [n,k]

    if cacu_rate:
        return [(test_labels == maxk[:, 0:i]).sum().item() / total for i in k]
    else:
        return total, [(test_labels == maxk[:, 0:i]).sum().item() for i in k]



from typing import TypeVar, Callable

T = TypeVar('T', Callable, str)

from functools import wraps


def clswrap(callable: T) -> T:
    """
    带类型提示的 staticmethod()
    Args:
        callable:

    Returns:
    Notes:
        必须在类中用
    """

    @staticmethod
    def inner(*args, **kwargs):
        return callable(*args, **kwargs)

    return inner



from typing import List, Tuple

import torch
from torch._utils import _accumulate
from torch.utils.data import Dataset, Subset


def sequence_split(lengths: List[int]) -> List[Dataset]:
    """
    对应于 torch.utils.data.dataset.random_split ，用于按照长度顺序切分数据集
    Args:
        lengths:

    Returns:

    """
    indices = torch.arange(0, sum(lengths)).tolist()
    return [indices[offset - length:offset] for offset, length in
            zip(_accumulate(lengths), lengths)]


def semi_split(labels, n_percls, val_size=10000, include_sup=True, repeat_sup=True, shuffle=True):
    """
    在半监督的情况下切分训练集，有标签训练集和无标签训练集
    """
    import numpy as np
    labels = np.array(labels)
    n_cls = len({int(i) for i in labels})

    n_per_un = (len(labels) - val_size) // n_cls
    indexs = []
    un_indexs = []
    val_indexs = []
    for i in range(n_cls):
        idx = np.where(labels == i)[0]

        np.random.shuffle(idx)
        indexs.extend(idx[:n_percls])

        if include_sup:
            un_indexs.extend(idx[:n_per_un])  # 无标签样本部份也使用有标签样板
        else:
            un_indexs.extend(idx[n_percls:n_per_un])

        val_indexs.extend(idx[n_per_un:])

    if repeat_sup:
        indexs = np.hstack([indexs for i in range((len(un_indexs) // len(indexs)) + 1)])
        indexs = indexs[:len(un_indexs)]

    if shuffle:
        np.random.shuffle(indexs)
        np.random.shuffle(un_indexs)
    return indexs, un_indexs, val_indexs


def train_val_split(target, val_size=10000, train_size=None):
    import numpy as np
    size = len(target)
    idx = np.arange(size)
    if isinstance(target,list):
        from lumo.base_classes import llist
        target = llist(target)

    idx = np.arange(len(target))
    np.random.shuffle(idx)

    if train_size is not None:
        assert size > val_size + train_size, "should less than {}, but {}".format(size, train_size + val_size)
        return idx[val_size:val_size + train_size], idx[:val_size]

    return idx[val_size:], idx[:val_size]


def ratio2length(total_len, *ratios) -> List[int]:
    """
    将比率转换成具体整数值
    Args:
        total_len:  总长度
        *ratios:  任意长度比率，和不要求为1

    Returns:
        List[int]

    """
    return [int(total_len * i) for i in ratios]



import os
import pickle as _pickle
from io import FileIO
from contextlib import contextmanager


def dump(obj, file, make_path=True, protocol=None, *, fix_imports=True):
    if isinstance(file, str):
        if make_path:
            os.makedirs(os.path.dirname(os.path.abspath(file)), exist_ok=True)
        file = open(file, 'wb')
        _pickle.dump(obj, file, protocol=protocol, fix_imports=fix_imports)
        file.close()
    elif isinstance(file, FileIO):
        _pickle.dump(obj, file, protocol=protocol, fix_imports=fix_imports)
    else:
        raise NotImplementedError()


def load(file, *, fix_imports=True, encoding="ASCII", errors="strict"):
    if isinstance(file, str):
        file = open(file, 'rb')
        res = _pickle.load(file, fix_imports=fix_imports, encoding=encoding, errors=errors)
        file.close()
        return res
    elif isinstance(file, FileIO):
        return _pickle.load(file, fix_imports=fix_imports, encoding=encoding, errors=errors)
    else:
        raise NotImplementedError()


def loads(s, *, fix_imports=True, encoding="ASCII", errors="strict"):
    return _pickle.loads(s, fix_imports=fix_imports, encoding=encoding, errors=errors)


def dumps(obj, protocol=None, *, fix_imports=True):
    return _pickle.dumps(obj, protocol=protocol, fix_imports=fix_imports)


@contextmanager
def cached(fn):
    import shutil
    cache_fn = f'{fn}.lumo_cache'
    try:
        yield cache_fn
    except:
        os.remove(cache_fn)
    finally:
        shutil.move(cache_fn, fn)



"""

"""

from lumo.utils.markdown import Markdown

if __name__ == '__main__':
    md = Markdown()
    md.add_title("report of exp")
    md.add_title("All loss report")
    md.add_table([[1, 2, 3], [4, 5, 6]])
    md.add_text("Today alisjoialn")
    md.add_url("www.baidu.com", "baidu")
    md.add_bold("aoisjdoij")
    md.add_picture("ww.bia", "ada", True)

    md.add_code("""
    print(asdas)
    for i in range():
        print(1)
    """)

    print(md)



import shutil
import sys
import os


def get_consolo_width():
    return shutil.get_terminal_size().columns - 1  # -1 for windows consolo


def support_multiline():
    if "jupyter_core" in sys.modules or shutil.get_terminal_size((0, 0)).columns == 0 or "PYCHARM_HOSTED" in os.environ:
        return True
    else:
        return False



import argparse
import os
import random
import shutil
import time
import warnings

import torch
import torch.nn as nn
import torch.nn.parallel
import torch.backends.cudnn as cudnn
import torch.distributed as dist
import torch.optim
import torch.multiprocessing as mp
import torch.utils.data
import torch.utils.data.distributed
import torchvision.transforms as transforms
import torchvision.datasets as datasets
import torchvision.models as models

model_names = sorted(name for name in models.__dict__
    if name.islower() and not name.startswith("__")
    and callable(models.__dict__[name]))

parser = argparse.ArgumentParser(description='PyTorch ImageNet Training')
parser.add_argument('data', metavar='DIR',
                    help='path to dataset')
parser.add_argument('-a', '--arch', metavar='ARCH', default='resnet18',
                    choices=model_names,
                    help='model architecture: ' +
                        ' | '.join(model_names) +
                        ' (default: resnet18)')
parser.add_argument('-j', '--workers', default=4, type=int, metavar='N',
                    help='number of data loading workers (default: 4)')
parser.add_argument('--epochs', default=90, type=int, metavar='N',
                    help='number of total epochs to run')
parser.add_argument('--start-epoch', default=0, type=int, metavar='N',
                    help='manual epoch number (useful on restarts)')
parser.add_argument('-b', '--batch-size', default=256, type=int,
                    metavar='N',
                    help='mini-batch size (default: 256), this is the total '
                         'batch size of all GPUs on the current node when '
                         'using Data Parallel or Distributed Data Parallel')
parser.add_argument('--lr', '--learning-rate', default=0.1, type=float,
                    metavar='LR', help='initial learning rate', dest='lr')
parser.add_argument('--momentum', default=0.9, type=float, metavar='M',
                    help='momentum')
parser.add_argument('--wd', '--weight-decay', default=1e-4, type=float,
                    metavar='W', help='weight decay (default: 1e-4)',
                    dest='weight_decay')
parser.add_argument('-p', '--print-freq', default=10, type=int,
                    metavar='N', help='print frequency (default: 10)')
parser.add_argument('--resume', default='', type=str, metavar='PATH',
                    help='path to latest checkpoint (default: none)')
parser.add_argument('-e', '--evaluate', dest='evaluate', action='store_true',
                    help='evaluate model on validation set')
parser.add_argument('--pretrained', dest='pretrained', action='store_true',
                    help='use pre-trained model')
parser.add_argument('--world-size', default=-1, type=int,
                    help='number of nodes for distributed training')
parser.add_argument('--rank', default=-1, type=int,
                    help='node rank for distributed training')
parser.add_argument('--dist-url', default='tcp://224.66.41.62:23456', type=str,
                    help='url used to set up distributed training')
parser.add_argument('--dist-backend', default='nccl', type=str,
                    help='distributed backend')
parser.add_argument('--seed', default=None, type=int,
                    help='seed for initializing training. ')
parser.add_argument('--gpu', default=None, type=int,
                    help='GPU id to use.')
parser.add_argument('--multiprocessing-distributed', action='store_true',
                    help='Use multi-processing distributed training to launch '
                         'N processes per node, which has N GPUs. This is the '
                         'fastest way to use PyTorch for either single node or '
                         'multi node data parallel training')

best_acc1 = 0


def main():
    args = parser.parse_args()

    if args.seed is not None:
        random.seed(args.seed)
        torch.manual_seed(args.seed)
        cudnn.deterministic = True
        warnings.warn('You have chosen to seed training. '
                      'This will turn on the CUDNN deterministic setting, '
                      'which can slow down your training considerably! '
                      'You may see unexpected behavior when restarting '
                      'from checkpoints.')

    if args.gpu is not None:
        warnings.warn('You have chosen a specific GPU. This will completely '
                      'disable data parallelism.')

    if args.dist_url == "env://" and args.world_size == -1:
        args.world_size = int(os.environ["WORLD_SIZE"])

    args.distributed = args.world_size > 1 or args.multiprocessing_distributed

    ngpus_per_node = torch.cuda.device_count()
    if args.multiprocessing_distributed:
        # Since we have ngpus_per_node processes per node, the total world_size
        # needs to be adjusted accordingly
        args.world_size = ngpus_per_node * args.world_size
        # Use torch.multiprocessing.spawn to launch distributed processes: the
        # main_worker process function
        mp.spawn(main_worker, nprocs=ngpus_per_node, args=(ngpus_per_node, args))
    else:
        # Simply call main_worker function
        main_worker(args.gpu, ngpus_per_node, args)


def main_worker(gpu, ngpus_per_node, args):
    global best_acc1
    args.gpu = gpu

    if args.gpu is not None:
        print("Use GPU: {} for training".format(args.gpu))

    if args.distributed:
        if args.dist_url == "env://" and args.rank == -1:
            args.rank = int(os.environ["RANK"])
        if args.multiprocessing_distributed:
            # For multiprocessing distributed training, rank needs to be the
            # global rank among all the processes
            args.rank = args.rank * ngpus_per_node + gpu
        dist.init_process_group(backend=args.dist_backend, init_method=args.dist_url,
                                world_size=args.world_size, rank=args.rank)
    # create model
    if args.pretrained:
        print("=> using pre-trained model '{}'".format(args.arch))
        model = models.__dict__[args.arch](pretrained=True)
    else:
        print("=> creating model '{}'".format(args.arch))
        model = models.__dict__[args.arch]()

    if not torch.cuda.is_available():
        print('using CPU, this will be slow')
    elif args.distributed:
        # For multiprocessing distributed, DistributedDataParallel constructor
        # should always set the single device scope, otherwise,
        # DistributedDataParallel will use all available devices.
        if args.gpu is not None:
            torch.cuda.set_device(args.gpu)
            model.cuda(args.gpu)
            # When using a single GPU per process and per
            # DistributedDataParallel, we need to divide the batch size
            # ourselves based on the total number of GPUs we have
            args.batch_size = int(args.batch_size / ngpus_per_node)
            args.workers = int((args.workers + ngpus_per_node - 1) / ngpus_per_node)
            model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[args.gpu])
        else:
            model.cuda()
            # DistributedDataParallel will divide and allocate batch_size to all
            # available GPUs if device_ids are not set
            model = torch.nn.parallel.DistributedDataParallel(model)
    elif args.gpu is not None:
        torch.cuda.set_device(args.gpu)
        model = model.cuda(args.gpu)
    else:
        # DataParallel will divide and allocate batch_size to all available GPUs
        if args.arch.startswith('alexnet') or args.arch.startswith('vgg'):
            model.features = torch.nn.DataParallel(model.features)
            model.cuda()
        else:
            model = torch.nn.DataParallel(model).cuda()

    # define loss function (criterion) and optimizer
    criterion = nn.CrossEntropyLoss().cuda(args.gpu)

    optimizer = torch.optim.SGD(model.parameters(), args.lr,
                                momentum=args.momentum,
                                weight_decay=args.weight_decay)

    # optionally resume from a checkpoint
    if args.resume:
        if os.path.isfile(args.resume):
            print("=> loading checkpoint '{}'".format(args.resume))
            if args.gpu is None:
                checkpoint = torch.load(args.resume)
            else:
                # Map model to be loaded to specified single gpu.
                loc = 'cuda:{}'.format(args.gpu)
                checkpoint = torch.load(args.resume, map_location=loc)
            args.start_epoch = checkpoint['epoch']
            best_acc1 = checkpoint['best_acc1']
            if args.gpu is not None:
                # best_acc1 may be from a checkpoint from a different GPU
                best_acc1 = best_acc1.to(args.gpu)
            model.load_state_dict(checkpoint['state_dict'])
            optimizer.load_state_dict(checkpoint['optimizer'])
            print("=> loaded checkpoint '{}' (epoch {})"
                  .format(args.resume, checkpoint['epoch']))
        else:
            print("=> no checkpoint found at '{}'".format(args.resume))

    cudnn.benchmark = True

    # Data loading code
    traindir = os.path.join(args.data, 'train')
    valdir = os.path.join(args.data, 'val')
    normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                     std=[0.229, 0.224, 0.225])

    train_dataset = datasets.ImageFolder(
        traindir,
        transforms.Compose([
            transforms.RandomResizedCrop(224),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            normalize,
        ]))

    if args.distributed:
        train_sampler = torch.utils.data.distributed.DistributedSampler(train_dataset)
    else:
        train_sampler = None

    train_loader = torch.utils.data.DataLoader(
        train_dataset, batch_size=args.batch_size, shuffle=(train_sampler is None),
        num_workers=args.workers, pin_memory=True, sampler=train_sampler)

    val_loader = torch.utils.data.DataLoader(
        datasets.ImageFolder(valdir, transforms.Compose([
            transforms.Resize(256),
            transforms.CenterCrop(224),
            transforms.ToTensor(),
            normalize,
        ])),
        batch_size=args.batch_size, shuffle=False,
        num_workers=args.workers, pin_memory=True)

    if args.evaluate:
        validate(val_loader, model, criterion, args)
        return

    for epoch in range(args.start_epoch, args.epochs):
        if args.distributed:
            train_sampler.set_epoch(epoch)
        adjust_learning_rate(optimizer, epoch, args)

        # train for one epoch
        train(train_loader, model, criterion, optimizer, epoch, args)

        # evaluate on validation set
        acc1 = validate(val_loader, model, criterion, args)

        # remember best acc@1 and save checkpoint
        is_best = acc1 > best_acc1
        best_acc1 = max(acc1, best_acc1)

        if not args.multiprocessing_distributed or (args.multiprocessing_distributed
                and args.rank % ngpus_per_node == 0):
            save_checkpoint({
                'epoch': epoch + 1,
                'arch': args.arch,
                'state_dict': model.state_dict(),
                'best_acc1': best_acc1,
                'optimizer' : optimizer.state_dict(),
            }, is_best)


def train(train_loader, model, criterion, optimizer, epoch, args):
    batch_time = AverageMeter('Time', ':6.3f')
    data_time = AverageMeter('Data', ':6.3f')
    losses = AverageMeter('Loss', ':.4e')
    top1 = AverageMeter('Acc@1', ':6.2f')
    top5 = AverageMeter('Acc@5', ':6.2f')
    progress = ProgressMeter(
        len(train_loader),
        [batch_time, data_time, losses, top1, top5],
        prefix="Epoch: [{}]".format(epoch))

    # switch to train mode
    model.train()

    end = time.time()
    for i, (images, target) in enumerate(train_loader):
        # measure data loading time
        data_time.update(time.time() - end)

        if args.gpu is not None:
            images = images.cuda(args.gpu, non_blocking=True)
        if torch.cuda.is_available():
            target = target.cuda(args.gpu, non_blocking=True)

        # compute output
        output = model(images)
        loss = criterion(output, target)

        # measure accuracy and record loss
        acc1, acc5 = accuracy(output, target, topk=(1, 5))
        losses.update(loss.item(), images.size(0))
        top1.update(acc1[0], images.size(0))
        top5.update(acc5[0], images.size(0))

        # compute gradient and do SGD step
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        if i % args.print_freq == 0:
            progress.display(i)


def validate(val_loader, model, criterion, args):
    batch_time = AverageMeter('Time', ':6.3f')
    losses = AverageMeter('Loss', ':.4e')
    top1 = AverageMeter('Acc@1', ':6.2f')
    top5 = AverageMeter('Acc@5', ':6.2f')
    progress = ProgressMeter(
        len(val_loader),
        [batch_time, losses, top1, top5],
        prefix='Test: ')

    # switch to evaluate mode
    model.eval()

    with torch.no_grad():
        end = time.time()
        for i, (images, target) in enumerate(val_loader):
            if args.gpu is not None:
                images = images.cuda(args.gpu, non_blocking=True)
            if torch.cuda.is_available():
                target = target.cuda(args.gpu, non_blocking=True)

            # compute output
            output = model(images)
            loss = criterion(output, target)

            # measure accuracy and record loss
            acc1, acc5 = accuracy(output, target, topk=(1, 5))
            losses.update(loss.item(), images.size(0))
            top1.update(acc1[0], images.size(0))
            top5.update(acc5[0], images.size(0))

            # measure elapsed time
            batch_time.update(time.time() - end)
            end = time.time()

            if i % args.print_freq == 0:
                progress.display(i)

        # TODO: this should also be done with the ProgressMeter
        print(' * Acc@1 {top1.avg:.3f} Acc@5 {top5.avg:.3f}'
              .format(top1=top1, top5=top5))

    return top1.avg


def save_checkpoint(state, is_best, filename='checkpoint.pth.tar'):
    torch.save(state, filename)
    if is_best:
        shutil.copyfile(filename, 'model_best.pth.tar')


class AverageMeter(object):
    """Computes and stores the average and current value"""
    def __init__(self, name, fmt=':f'):
        self.name = name
        self.fmt = fmt
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = '{name} {val' + self.fmt + '} ({avg' + self.fmt + '})'
        return fmtstr.format(**self.__dict__)


class ProgressMeter(object):
    def __init__(self, num_batches, meters, prefix=""):
        self.batch_fmtstr = self._get_batch_fmtstr(num_batches)
        self.meters = meters
        self.prefix = prefix

    def display(self, batch):
        entries = [self.prefix + self.batch_fmtstr.format(batch)]
        entries += [str(meter) for meter in self.meters]
        print('\t'.join(entries))

    def _get_batch_fmtstr(self, num_batches):
        num_digits = len(str(num_batches // 1))
        fmt = '{:' + str(num_digits) + 'd}'
        return '[' + fmt + '/' + fmt.format(num_batches) + ']'


def adjust_learning_rate(optimizer, epoch, args):
    """Sets the learning rate to the initial LR decayed by 10 every 30 epochs"""
    lr = args.lr * (0.1 ** (epoch // 30))
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr


def accuracy(output, target, topk=(1,)):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    with torch.no_grad():
        maxk = max(topk)
        batch_size = target.size(0)

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.view(1, -1).expand_as(pred))

        res = []
        for k in topk:
            correct_k = correct[:k].reshape(-1).float().sum(0, keepdim=True)
            res.append(correct_k.mul_(100.0 / batch_size))
        return res


if __name__ == '__main__':
    main()


"""
"""
import operator as op

import numpy as np


class Constrain():
    def __init__(self):
        self._name = None

    def __getattr__(self, item: str):
        if item.startswith('_'):
            return AttributeError(item)
        else:
            self._name = item
            return self

    def __getitem__(self, item: str):
        if item.startswith('_'):
            return AttributeError(item)
        self._name = item
        return self


class MeterConstrain(Constrain):
    """
    A meter constrain, mainly used for Boardquery.parallel()

    Examples:
    ```python
    from thexp import Q,C
    query = Q.tests().boards()
    print(query.scalar_tags)
    query.parallel(C.param.lr,C.param.epoch,C.meter.loss_train_,C.meter.acc_test_)
    ```
    """

    def __init__(self):
        super().__init__()
        self._constrain = None

    @property
    def max(self):
        self._constrain = np.max
        return self

    @property
    def min(self):
        self._constrain = np.min
        return self

    def __repr__(self):
        if self._constrain is not None:
            name = self._constrain.__name__.lstrip('a')
        else:
            name = '<empty>'

        return "Meter.{}({})".format(name, self._name)


_op_str = {
    'lt': '< ',
    'gt': '> ',
    'le': '<=',
    'ge': '>=',
    'eq': '==',
    'ne': '!=',
}


class ParamConstrain(Constrain):
    """
    A param constrain object corresponds to a constrain of a hyperparameter in Params class, which have a name,
    a comparison operator, a flag(whether can be none), and the comparison value.

    You can create this constrain by class ConstrainBuilder, which have a global instance named `C`

    >>> from thexp import C
    >>> constrain = (C.param.dataset == 'cifar10')
    ... print(constrain)

    Param(dataset == cifar10)


    >>> constrain = (C.param.allow_none.margin > 5)
    ... print(constrain)

    Param(Optional| margin >  5)


    You can use this instance in BoardQuery.parallel() to declare param name, or in TestQuery.filter_params() to query tests.

    Examples:
    ---------
    from thexp import Q
    Q.tests().filter_params(
        C.param.margin > 5,
        C.param.allow_none.margin > 5,
        C.param.epoch > 100,
        C.param.dataset == 'cifar10',
    )

    from thexp import Q
    Q.tests().boards().parallel(
        C.param.lr,
        C.meter.loss_train_,
        C.param.dataset
    )
    # for simpleness, you can create name with C.param_names('lr','margin',...)
    Q.tests().boards().parallel(
        *C.param_names('lr','margin'),
    )

    """

    def __init__(self):
        super().__init__()
        self._constrain = None
        self._allow_none = False
        self._value = None

    def __lt__(self, other):
        self._constrain = op.lt
        self._value = other
        return self

    def lt(self, other):
        self._constrain = op.lt
        self._value = other
        return self

    def __gt__(self, other):
        self._constrain = op.gt
        self._value = other
        return self

    def gt(self, other):
        self._constrain = op.gt
        self._value = other
        return self

    def __le__(self, other):
        self._constrain = op.le
        self._value = other
        return self

    def le(self, other):
        self._constrain = op.le
        self._value = other
        return self

    def __ge__(self, other):
        self._constrain = op.ge
        self._value = other
        return self

    def ge(self, other):
        self._constrain = op.ge
        self._value = other
        return self

    def __eq__(self, other):
        self._constrain = op.eq
        self._value = other
        return self

    def eq(self, other):
        self._constrain = op.eq
        self._value = other
        return self

    def __ne__(self, other):
        self._constrain = op.ne
        self._value = other
        return self

    def ne(self, other):
        self._constrain = op.ne
        self._value = other
        return self

    def __repr__(self):
        if self._allow_none:
            off = 'Optional| '
        else:
            off = ''

        return 'Param({}{} {} {})'.format(off, self._name, _op_str[self._constrain.__name__], self._value)

    @property
    def allow_none(self):
        self._allow_none = True
        return self


class ConstrainBuilder():
    @property
    def param(self):
        return ParamConstrain()

    @property
    def meter(self):
        return MeterConstrain()

    def param_names(self, *names: str):
        return [ParamConstrain()[name] for name in names]

    def meter_names(self, *names: str):
        return [MeterConstrain()[name] for name in names]


C = ConstrainBuilder()



"""
Used for recording data
"""
from collections import OrderedDict
from numbers import Number
from typing import Any, Union

import numpy as np
import torch

from ..base_classes.errors import MeterNameException
from ..base_classes.trickitems import NoneItem


class _Format:
    def __call__(self, value):
        raise NotImplementedError()


class _IntFmt(_Format):
    def __call__(self, value):
        return f"{value:.0f}"


class _FloatFmt(_Format):
    def __init__(self, precision=2):
        self.prec = precision

    def __call__(self, value):
        return f"{value:.{self.prec}f}"


class _PercentFmt(_FloatFmt):
    def __call__(self, value):
        return f"{value:.{self.prec}%}"


class _TensorFmt(_FloatFmt):

    def __call__(self, value):
        if len(value.shape) != 0:
            value = value.item()
        return f"{value:.{self.prec}f}"


class _InlineStrFmt(_Format):

    def __call__(self, value):
        l = str(value).split("\n")
        if len(l) > 1:
            return f"{l[0]}..."
        else:
            return l[0]


class _AvgItem:
    """
    用于保存累积均值的类
    avg = AvgItem()
    avg += 1 # avg.update(1)
    avg += 2
    avg += 3

    avg.item = 3 #(last item)
    avg.avg = 2 #(average item)
    avg.sum = 6
    """

    def __init__(self, weight_decay=0.01, precision=4) -> None:
        super().__init__()
        self._sum = None
        self._count = 0
        self._last = None
        self._w = weight_decay
        self._p = precision

    def update(self, other):
        if self._sum is None:
            self._sum = other
        else:
            self._sum = self._sum + other
        self._count += 1
        self._last = other

    @property
    def last(self) -> Number:
        if self._last is None:
            return 0
        return self._last

    @property
    def avg(self) -> Number:
        if self._sum is None:
            return 0
        return self._sum / self._count

    def __repr__(self) -> str:
        return f"{self.avg:.{self._p}f}"

    def __format__(self, format_spec):
        return f"{self.avg}:{format_spec}"


class Meter:
    """

    Examples:
    --------
    m = Meter()
    m.short()
    m.int()
    m.float()
    m.percent()

    m.k += 10
    m.k.update()

    m.k.backstep()
    """

    def __init__(self, precision=4):
        self._prec = precision
        self._meter_dict = OrderedDict()
        self._format_dict = dict()
        self._convert_type = []

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            super().__setattr__(name, value)
        elif name.endswith("_"):
            raise MeterNameException(f"Meter name can't end with '_', but got '{name}'")
        else:
            value = self._convert(value)

            if isinstance(value, int) and name not in self._format_dict:
                self.int(name)
            elif isinstance(value, float) and name not in self._format_dict:
                self.float(name)
            elif isinstance(value, torch.Tensor) and name not in self._format_dict:
                if len(value.shape) == 0 or (sum(value.shape) == 1):
                    self.tensorfloat(name)
                else:
                    self.str_in_line(name)

            self._meter_dict[name] = value

    def __getattr__(self, item):
        if item.endswith("_"):
            return item.rstrip("_")

        if item not in self._meter_dict:
            return NoneItem()
        else:
            return self._meter_dict[item]

    def __getitem__(self, item):
        item = str(item)
        return self.__getattr__(item)

    def __setitem__(self, key, value):
        key = str(key)
        self.__setattr__(key, value)

    def __iter__(self):
        return iter(self._meter_dict)

    def __repr__(self):
        log_dict = self.serialize()
        return " | ".join(["{}: {}".format(k, v) for k, v in log_dict.items()])

    def _convert(self, val):
        if type(val) in {int, float, bool, str}:
            return val
        for tp, func in self._convert_type:
            if type(val) == tp:
                val = func(val)
        return val

    def serialize(self) -> OrderedDict:
        """format all value to be strings to make sure it can be printed without Exception"""
        log_dict = OrderedDict()
        for k, v in self._meter_dict.items():
            if k in self._format_dict:
                v = self._format_dict[k](v)
            log_dict[k] = v
        return log_dict

    def int(self, key: str):
        """
        Variable with name `key` will be formatted as int when log.
        Args:
            key: Variable name
        """
        self._format_dict[key] = _IntFmt()
        return self

    def float(self, key: str, precision=3):
        """
        Variable with name `key` will be formatted as float when log.

        Args:
            key: Variable name
            precision:
        """
        self._format_dict[key] = _FloatFmt(precision)
        return self

    def percent(self, key: str, precision=2):
        """
        Variable with name `key` will be formatted as a percentage when log.
        Args:
            key: Variable name
            precision:
        """
        self._format_dict[key] = _PercentFmt(precision)
        return self

    def tensorfloat(self, key: str, precision=4):
        """
        Format torch.Tensor object as a float when log.
        Args:
            key: Variable name.
            precision:
        """
        self._format_dict[key] = _TensorFmt(precision)
        return self

    def str_in_line(self, key: str):
        self._format_dict[key] = _InlineStrFmt()
        return self

    def add_format_type(self, type, func):
        self._convert_type.append((type, func))
        return self

    def items(self):
        """like dict.items()"""
        for k, v in self._meter_dict.items():
            yield k, v

    def array_items(self):
        """
        return items that is a Number/torch.Tensor/np.ndarray object
        Returns:
            iterated (k, v) sequence, where v is an instance of Number/torch.Tensor/np.ndarray
        """
        for k, v in self._meter_dict.items():
            if isinstance(v, (Number, torch.Tensor, np.ndarray)):
                yield k, v

    def numeral_items(self):
        """
        return items that can be converted into one scalar.
        Returns:
            iterated (k, v) sequence, where v is int or float object
        Notes:
            all array(torch.Tensor/np.ndarray) that only has one element will be yiled, either.
        """
        for k, v in self._meter_dict.items():
            if isinstance(v, (int, float)):
                yield k, v
            elif isinstance(v, torch.Tensor):
                try:
                    yield k, v.detach().cpu().item()
                except ValueError:  # not an one element tensor
                    continue
            elif isinstance(v, np.ndarray):
                try:
                    yield k, v.item()
                except ValueError:  # not an one element tensor
                    continue

    def update(self, meter: dict):
        """like dict.update()"""
        for k, v in meter.items():
            self[k] = v
        return self


class AvgMeter(Meter):
    """

    """

    def __init__(self, weight_decay=0.75, precision=4):
        super().__init__(precision)
        self._weight_decay = weight_decay

    def __getitem__(self, item) -> _AvgItem:
        return super().__getitem__(item)

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            super().__setattr__(name, value)
        elif name.endswith("_"):
            raise MeterNameException(f"Meter name can't end with '_', but got '{name}'")
        else:
            value = self._convert(value)
            if isinstance(value, int) and name not in self._format_dict:
                self.int(name)
            elif isinstance(value, float) and name not in self._format_dict:
                self.float(name)
            elif isinstance(value, torch.Tensor) and name not in self._format_dict:
                if len(value.shape) == 0 or (sum(value.shape) == 1):
                    self.tensorfloat(name)
                else:
                    self.str_in_line(name)

            if isinstance(value, (torch.Tensor, np.ndarray)):
                if isinstance(value, torch.Tensor):
                    value = value.detach().cpu().numpy()
                if len(value.shape) == 0 or sum(value.shape) == 1:
                    if name not in self._meter_dict:
                        self._meter_dict[name] = self.avg_item()
                    self._meter_dict[name].update(value)
                else:
                    self._meter_dict[name] = value
            elif name in self._meter_dict and isinstance(self._meter_dict[name], _AvgItem):
                self._meter_dict[name].update(value)
            elif isinstance(value, (float)):
                if name not in self._meter_dict:
                    self._meter_dict[name] = self.avg_item()
                self._meter_dict[name].update(value)
            else:
                self._meter_dict[name] = value

    def avg_item(self):
        return _AvgItem(weight_decay=0.75, precision=self._prec)

    def update(self, meter: Union[Meter, dict]):
        if meter is None:
            return
        for k, v in meter.items():
            self[k] = v
        if isinstance(meter, Meter):
            self._format_dict.update(meter._format_dict)
            self._convert_type.extend(meter._convert_type)

    def serialize(self):
        log_dict = OrderedDict()
        for k, v in self._meter_dict.items():
            if k in self._format_dict:
                if isinstance(v, _AvgItem):
                    v = v.avg
                v = self._format_dict[k](v)
            log_dict[k] = v
        return log_dict

    def array_items(self):
        for k, v in self._meter_dict.items():
            if isinstance(v, (int, float, torch.Tensor, np.ndarray)):
                yield k, v
            elif isinstance(v, _AvgItem):
                yield k, v.avg

    def numeral_items(self):
        """"""
        for k, v in self._meter_dict.items():
            if isinstance(v, (int, float)):
                yield k, v
            elif isinstance(v, torch.Tensor):
                try:
                    yield k, v.detach().cpu().item()
                except:
                    continue
            elif isinstance(v, np.ndarray):
                try:
                    yield k, v.item()
                except:
                    continue
            elif isinstance(v, _AvgItem):
                yield k, v.avg

    @property
    def meter(self):
        """try to get newest value"""
        return _InMeter(self)

    @property
    def avg(self):
        """try to get average value"""
        return _InMeter(self, mode='avg')


class _InMeter():
    def __init__(self, avgmeter: AvgMeter, mode='newest'):
        self._avg = avgmeter
        self._mode = mode

    def __getattr__(self, item):
        if item not in self._avg._meter_dict:
            raise AttributeError(item)
        res = self._avg._meter_dict[item]
        if isinstance(res, _AvgItem):
            if self._mode == 'newest':
                res = res.last
            else:
                res = res.avg
        return res

    def __getitem__(self, item: str):
        return self.__getattr__(item)
